# ruff: noqa: FBT003
from __future__ import annotations

from typing import TYPE_CHECKING

from deltona.media import CD_FRAMES
from deltona.utils import (
    DataAdapter,
    add_cdda_times,
    kill_processes_by_name,
    secure_move_path,
    unregister_wine_file_associations,
)
import pytest
import requests

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch
    from pytest_mock import MockerFixture


def test_add_cdda_times_none_or_empty() -> None:
    assert add_cdda_times(None) is None
    assert add_cdda_times([]) is None


def test_add_cdda_times_invalid_format() -> None:
    # Not matching MM:SS:FF
    assert add_cdda_times(['12:34']) is None
    assert add_cdda_times(['99:99:99']) is None
    assert add_cdda_times(['abc:def:ghi']) is None
    assert add_cdda_times(['12:34:']) is None


def test_add_cdda_times_valid_single() -> None:
    # 01:02:03 should be valid
    result = add_cdda_times(['01:02:03'])
    assert isinstance(result, str)
    assert result.count(':') == 2


def test_add_cdda_times_valid_multiple() -> None:
    result = add_cdda_times(['00:01:00', '00:01:00'])
    assert result.startswith('00:02:')  # type: ignore[union-attr]


def test_add_cdda_times_overflow_minutes() -> None:
    # Minutes > 99 should return None
    # Use a time that will sum to > 99 minutes
    times = ['99:59:74', '00:01:01']
    assert add_cdda_times(times) is None


def test_add_cdda_times_overflow_seconds() -> None:
    # Seconds > 59 should return None
    times = ['00:60:00']
    assert add_cdda_times(times) is None


def test_add_cdda_times_overflow_frames() -> None:
    # Frames > CD_FRAMES should return None
    times = [f'00:00:{CD_FRAMES + 1:02d}']
    assert add_cdda_times(times) is None


def test_add_cdda_times_exact_maximum() -> None:
    # Exactly at the maximum allowed
    times = ['99:59:74']
    result = add_cdda_times(times)
    assert isinstance(result, str)
    assert result.startswith('99:59:')


def test_add_cdda_times_leading_zeros() -> None:
    # Should handle leading zeros
    result = add_cdda_times(['00:00:01', '00:00:01'])
    assert result.startswith('00:00:')  # type: ignore[union-attr]


def test_unregister_wine_file_associations_basic(mocker: MockerFixture) -> None:
    kill_wine_mock = mocker.patch('deltona.utils.kill_wine')
    mock_file1 = mocker.Mock()
    mock_file2 = mocker.Mock()
    mock_file3 = mocker.Mock()
    mock_file4 = mocker.Mock()
    mock_file5 = mocker.Mock()
    mock_file6 = mocker.Mock()
    mock_file7 = mocker.Mock()
    mock_file8 = mocker.Mock()
    mimeinfo_mock = mocker.Mock()
    mock_path = mocker.patch('deltona.utils.Path')
    mock_path.home.return_value.__truediv__.return_value = mimeinfo_mock
    mock_path.home.return_value.__truediv__.return_value.glob.side_effect = [
        [mock_file1, mock_file2],  # wine-extension-*.desktop
        [mock_file3, mock_file4],  # x-wine*
        [mock_file5, mock_file6],  # x-wine-extension*
    ]
    # application-x-wine-extension*
    mock_path.home.return_value.__truediv__.return_value.rglob.return_value = [
        mock_file7,
        mock_file8,
    ]
    sp_run = mocker.patch('deltona.utils.sp.run')
    unregister_wine_file_associations()
    kill_wine_mock.assert_called_once()
    for _i, f in enumerate(
        [
            mock_file1,
            mock_file2,
            mock_file3,
            mock_file4,
            mock_file5,
            mock_file6,
            mock_file7,
            mock_file8,
        ],
            start=1,
    ):
        f.unlink.assert_called()
    mimeinfo_mock.unlink.assert_called()
    assert sp_run.call_count >= 2


def test_unregister_wine_file_associations_debug_true(mocker: MockerFixture) -> None:
    kill_wine_mock = mocker.patch('deltona.utils.kill_wine')
    mocker.patch('deltona.utils.Path.glob', return_value=[])
    mocker.patch('deltona.utils.Path.rglob', return_value=[])
    mimeinfo_mock = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=mimeinfo_mock)
    mimeinfo_mock.unlink = mocker.Mock()
    sp_run = mocker.patch('deltona.utils.sp.run')
    unregister_wine_file_associations(debug=True)
    called_args = [call[0][0] for call in sp_run.call_args_list]
    assert any('update-desktop-database' in str(args) for args in called_args)
    assert any('update-mime-database' in str(args) for args in called_args)
    kill_wine_mock.assert_called_once()


def test_secure_move_path_file_basic(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = True
    path_instance.name = 'file.txt'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    mocker.patch('deltona.utils.os.walk')
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    path_instance.unlink = mocker.Mock()
    secure_move_path(client, 'file.txt', '~/target')
    assert sftp.put.called
    assert path_instance.unlink.called


def test_secure_move_path_file_basic_preserve_stats(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = True
    path_instance.name = 'file.txt'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    mocker.patch('deltona.utils.os.walk')
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    path_instance.unlink = mocker.Mock()
    secure_move_path(client, 'file.txt', '~/target', preserve_stats=True)
    assert sftp.put.called
    assert sftp.utime.called
    assert path_instance.unlink.called


def test_secure_move_path_file_dry_run(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = True
    path_instance.name = 'file.txt'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    path_instance.unlink = mocker.Mock()
    secure_move_path(client, 'file.txt', '~/target', dry_run=True)
    assert not sftp.put.called
    assert not path_instance.unlink.called


def test_secure_move_path_directory_basic(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch(
        'deltona.utils.os.walk',
        side_effect=[[('/src', ['subdir'], ['file1', 'file2'])], [('/src', ['subdir'], [])]],
    )
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    sftp.put = mocker.Mock()
    sftp.utime = mocker.Mock()
    sftp.remove = mocker.Mock()
    sftp.rmdir = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=path_instance)
    secure_move_path(client, '/src', '~/target', preserve_stats=True)
    assert sftp.mkdir.called
    assert sftp.put.called
    assert sftp.utime.called


def test_secure_move_path_preserve_stats(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = True
    path_instance.name = 'file.txt'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    sftp.utime = mocker.Mock()
    secure_move_path(client, 'file.txt', '~/target', preserve_stats=True)
    assert sftp.utime.called


def test_secure_move_path_write_into(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch('deltona.utils.os.walk', return_value=[('/src', [], ['file1'])])
    sftp.stat = mocker.Mock()
    secure_move_path(client, '/src', '~/target', write_into=True)
    assert sftp.stat.called


def test_secure_move_path_handles_file_not_found_on_stat(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch('deltona.utils.os.walk', return_value=[('/src', [], [])])
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    secure_move_path(client, '/src', '~/target')
    assert sftp.mkdir.called


def test_secure_move_path_file_dry_run_preserve_stats(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = True
    path_instance.name = 'file.txt'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    path_instance.unlink = mocker.Mock()
    sftp.utime = mocker.Mock()
    secure_move_path(client, 'file.txt', '~/target', dry_run=True, preserve_stats=True)
    assert not sftp.put.called
    assert not path_instance.unlink.called
    assert not sftp.utime.called


def test_secure_move_path_dir_dry_run_preserve_stats(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch(
        'deltona.utils.os.walk',
        side_effect=[[('/src', ['subdir'], ['file1', 'file2'])], [('/src', ['subdir'], [])]],
    )
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    sftp.put = mocker.Mock()
    sftp.utime = mocker.Mock()
    sftp.remove = mocker.Mock()
    sftp.rmdir = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=path_instance)
    secure_move_path(client, '/src', '~/target', dry_run=True, preserve_stats=True)
    assert not sftp.put.called
    assert not sftp.utime.called
    assert not sftp.remove.called
    assert not sftp.rmdir.called


def test_secure_move_path_dir_preserve_stats_alt(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch(
        'deltona.utils.os.walk',
        side_effect=[[('/src', ['subdir'], ['file1', 'file2'])], [('/src', ['subdir'], [])]],
    )
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    sftp.put = mocker.Mock()
    sftp.utime = mocker.Mock()
    sftp.remove = mocker.Mock()
    sftp.rmdir = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=path_instance)
    secure_move_path(client, '/src', '~/target', preserve_stats=True, write_into=True)
    assert sftp.put.called
    assert sftp.utime.called
    assert not sftp.remove.called
    assert not sftp.rmdir.called


def test_secure_move_path_dir_preserve_stats_alt_dry_run(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch(
        'deltona.utils.os.walk',
        side_effect=[[('/src', ['subdir'], ['file1', 'file2'])], [('/src', ['subdir'], [])]],
    )
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    sftp.put = mocker.Mock()
    sftp.utime = mocker.Mock()
    sftp.remove = mocker.Mock()
    sftp.rmdir = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=path_instance)
    secure_move_path(client, '/src', '~/target', preserve_stats=True, write_into=True, dry_run=True)
    assert not sftp.put.called
    assert not sftp.utime.called
    assert not sftp.remove.called
    assert not sftp.rmdir.called


def test_secure_move_path_dir_write_into_no_preserve_stats(mocker: MockerFixture) -> None:
    client = mocker.MagicMock()
    sftp = mocker.MagicMock()
    client.open_sftp.return_value.__enter__.return_value = sftp
    path_mock = mocker.patch('deltona.utils.Path')
    path_instance = path_mock.return_value
    path_instance.is_file.return_value = False
    path_instance.name = 'dir'
    path_instance.stat.return_value = mocker.Mock(st_atime=1.0, st_mtime=2.0)
    client.exec_command.return_value = (None, mocker.Mock(read=lambda: b'/home/remote'), None)
    mocker.patch(
        'deltona.utils.os.walk',
        side_effect=[[('/src', ['subdir'], ['file1', 'file2'])], [('/src', ['subdir'], [])]],
    )
    sftp.stat.side_effect = FileNotFoundError
    sftp.mkdir = mocker.Mock()
    sftp.put = mocker.Mock()
    sftp.utime = mocker.Mock()
    sftp.remove = mocker.Mock()
    sftp.rmdir = mocker.Mock()
    mocker.patch('deltona.utils.Path', return_value=path_instance)
    secure_move_path(client, '/src', '~/target', write_into=True)
    assert sftp.put.called
    assert not sftp.utime.called
    assert not sftp.remove.called
    assert not sftp.rmdir.called


def test_kill_processes_by_name_windows_basic(mocker: MockerFixture) -> None:
    mocker.patch('deltona.utils.IS_WINDOWS', True)
    run_mock = mocker.patch('deltona.utils.sp.run')
    result = kill_processes_by_name('notepad')
    run_mock.assert_any_call(('taskkill.exe', '/im', 'notepad.exe'),
                             check=False,
                             capture_output=True)
    assert result is None


def test_kill_processes_by_name_unix_basic(mocker: MockerFixture) -> None:
    mocker.patch('deltona.utils.IS_WINDOWS', False)
    run_mock = mocker.patch('deltona.utils.sp.run')
    result = kill_processes_by_name('bash')
    run_mock.assert_any_call(('killall', '-15', 'bash'), check=False, capture_output=True)
    assert result is None


def test_kill_processes_by_name_windows_with_wait_timeout(mocker: MockerFixture) -> None:
    mocker.patch('deltona.utils.IS_WINDOWS', True)
    run_mock = mocker.patch('deltona.utils.sp.run')
    run_mock.side_effect = [
        mocker.Mock(),  # taskkill
        mocker.Mock(
            stdout='"Image Name","PID"\n"notepad.exe","1234"\n"notepad.exe","5678"\n'),  # tasklist
    ]
    sleep_mock = mocker.patch('deltona.utils.time.sleep')
    result = kill_processes_by_name('notepad', wait_timeout=0.1)
    assert result == [1234, 5678]
    sleep_mock.assert_called_once_with(0.1)


def test_kill_processes_by_name_unix_with_wait_timeout_and_force(mocker: MockerFixture) -> None:
    mocker.patch('deltona.utils.IS_WINDOWS', False)
    run_mock = mocker.patch('deltona.utils.sp.run')
    run_mock.side_effect = [
        mocker.Mock(),  # killall
        mocker.Mock(stdout='1234 bash\n5678 bash\n9999 other\n'),  # ps
        mocker.Mock(stdout=''),  # kill -9
    ]
    sleep_mock = mocker.patch('deltona.utils.time.sleep')
    mocker.patch('deltona.utils.Path.name', new_callable=mocker.PropertyMock, return_value='bash')
    result = kill_processes_by_name('bash', wait_timeout=0.2, force=True)
    assert result == [1234, 5678, 9999]
    sleep_mock.assert_called_once_with(0.2)


def test_kill_processes_by_name_no_processes_left(mocker: MockerFixture,
                                                  monkeypatch: MonkeyPatch) -> None:
    mocker.patch('deltona.utils.IS_WINDOWS', True)
    run_mock = mocker.patch('deltona.utils.sp.run')
    run_mock.side_effect = [
        mocker.Mock(),  # taskkill
        mocker.Mock(stdout='"Image Name","PID"\n'),  # tasklist
    ]
    sleep_mock = mocker.patch('deltona.utils.time.sleep')
    result = kill_processes_by_name('notepad', wait_timeout=0.1)
    assert result == []
    sleep_mock.assert_not_called()


def test_data_adapter_send_basic(mocker: MockerFixture) -> None:
    adapter = DataAdapter()
    session = requests.Session()
    session.mount('data:', adapter)
    req = requests.Request('GET', 'data:,HelloWorld').prepare()
    response = adapter.send(req)
    assert response.status_code == 200
    assert response.content == b',HelloWorld'


def test_data_adapter_send_with_stream_and_timeout(mocker: MockerFixture) -> None:
    adapter = DataAdapter()
    req = requests.Request('GET', 'data:,TestData').prepare()
    response = adapter.send(req, stream=True, timeout=5)
    assert response.status_code == 200
    assert response.content == b',TestData'


def test_data_adapter_send_with_cert_and_proxies(mocker: MockerFixture) -> None:
    adapter = DataAdapter()
    req = requests.Request('GET', 'data:,CertTest').prepare()
    response = adapter.send(req, cert='dummy', proxies={'http': 'proxy'})
    assert response.status_code == 200
    assert response.content == b',CertTest'


def test_data_adapter_send_assert_url_none(mocker: MockerFixture) -> None:
    adapter = DataAdapter()
    req = mocker.Mock()
    req.url = None
    with pytest.raises(AssertionError):
        adapter.send(req)


def test_data_adapter_close_noop(mocker: MockerFixture) -> None:
    adapter = DataAdapter()
    adapter.close()



from __future__ import annotations

from typing import TYPE_CHECKING
import enum
import subprocess as sp

from click.testing import CliRunner
from deltona.commands.media import (
    add_cdda_times_main,
    add_info_json_main,
    audio2vid_main,
    cddb_query_main,
    display_info_json_main,
    encode_dashcam_main,
    flac_dir_finalize_main,
    hlg2sdr_main,
    ke_ebook_ex_main,
    mvid_rename_main,
    supported_audio_input_formats_main,
    tbc2srt_main,
    ultraiso_main,
    wait_for_disc_main,
)
from deltona.ultraiso import InsufficientArguments
from pytest_mock import MockerFixture
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


class _ReturnCodeType(enum.IntEnum):
    SUCCESS = enum.auto()
    FAILURE = enum.auto()


@pytest.mark.parametrize(
    ('input_', 'cdda_return', 'return_code_type'),
    [
        ('', None, _ReturnCodeType.FAILURE),
        ('01:23:45', None, _ReturnCodeType.FAILURE),
        ('01:23:45', '01:23:45', _ReturnCodeType.SUCCESS),
        ('00:00:00', '00:00:00', _ReturnCodeType.SUCCESS),
        ('12:34:00', '12:34:00', _ReturnCodeType.SUCCESS),
        ('01:02:03', '01:02:03', _ReturnCodeType.SUCCESS),
        ('01:02:00', '01:02:00', _ReturnCodeType.SUCCESS),
        ('00:01:', None, _ReturnCodeType.FAILURE),
    ],
)
def test_add_cdda_times_main_success(
    input_: str,
    cdda_return: str,
    return_code_type: _ReturnCodeType,
    mocker: MockerFixture,
    runner: CliRunner,
) -> None:
    mocker.patch('deltona.commands.media.add_cdda_times', return_value=cdda_return)
    result = runner.invoke(add_cdda_times_main, [input_])
    if return_code_type == _ReturnCodeType.SUCCESS:
        assert result.exit_code == 0
        assert cdda_return in result.output
    else:
        assert result.exit_code != 0


def test_wait_for_disc_main_success(mocker: MockerFixture, runner: CliRunner,
                                    tmp_path: Path) -> None:
    mocker.patch('deltona.commands.media.wait_for_disc', return_value=True)
    file = tmp_path / 'disc'
    file.write_text('dummy')
    result = runner.invoke(wait_for_disc_main, [str(file)])
    assert result.exit_code == 0


def test_wait_for_disc_main_fail(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.media.wait_for_disc', return_value=False)
    file = tmp_path / 'disc'
    file.write_text('dummy')
    result = runner.invoke(wait_for_disc_main, [str(file)])
    assert result.exit_code != 0


def test_ultraiso_main_success(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.media.run_ultraiso')
    result = runner.invoke(ultraiso_main, ['--appid', 'test', '--output', 'out.iso'])
    assert result.exit_code == 0


def test_ultraiso_main_insufficient_args(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.media.run_ultraiso', side_effect=InsufficientArguments)
    result = runner.invoke(ultraiso_main, ['--appid', 'test'])
    assert result.exit_code != 0


def test_ultraiso_main_file_not_found_error(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.media.run_ultraiso', side_effect=FileNotFoundError)
    result = runner.invoke(ultraiso_main, ['--appid', 'test'])
    assert result.exit_code != 0


def test_supported_audio_input_formats_main_success(mocker: MockerFixture,
                                                    runner: CliRunner) -> None:
    mocker.patch('deltona.commands.media.supported_audio_input_formats',
                 return_value=[('wav', 44100)])
    result = runner.invoke(supported_audio_input_formats_main, ['dummy'])
    assert result.exit_code == 0
    assert 'wav @ 44100' in result.output


def test_supported_audio_input_formats_main_oserror(mocker: MockerFixture,
                                                    runner: CliRunner) -> None:
    mocker.patch('deltona.commands.media.supported_audio_input_formats', side_effect=OSError)
    result = runner.invoke(supported_audio_input_formats_main, ['dummy'])
    assert result.exit_code != 0


def test_add_info_json_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    f = tmp_path / 'file.mp3'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.add_info_json_to_media_file')
    result = runner.invoke(add_info_json_main, [str(f)])
    assert result.exit_code == 0


def test_display_info_json_main_success(mocker: MockerFixture, runner: CliRunner,
                                        tmp_path: Path) -> None:
    f = tmp_path / 'file.mp3'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.get_info_json', return_value='{"foo": "bar"}')
    result = runner.invoke(display_info_json_main, [str(f)])
    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


def test_display_info_json_main_not_implemented(mocker: MockerFixture, runner: CliRunner,
                                                tmp_path: Path) -> None:
    f = tmp_path / 'file.mp3'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.get_info_json', side_effect=NotImplementedError)
    result = runner.invoke(display_info_json_main, [str(f)])
    assert result.exit_code != 0


def test_display_info_json_main_subprocess_error(mocker: MockerFixture, runner: CliRunner,
                                                 tmp_path: Path) -> None:
    f = tmp_path / 'file.mp3'
    f.write_text('dummy')
    mocker.patch(
        'deltona.commands.media.get_info_json',
        side_effect=sp.CalledProcessError(returncode=1, cmd='ffprobe', output='Error'),
    )
    result = runner.invoke(display_info_json_main, [str(f)])
    assert result.exit_code != 0


def test_audio2vid_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    f = tmp_path / 'audio.mp3'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.create_static_text_video')
    result = runner.invoke(audio2vid_main, [str(f), 'hello', 'world'])
    assert result.exit_code == 0


def test_mvid_rename_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    d = tmp_path / 'testdir'
    d.mkdir()
    src = d / 'testdir.mkv'
    src.write_text('dummy')
    non_dir = d / 'non_dir.mkv'
    non_dir.write_text('dummy')
    mocker.patch('deltona.commands.media.send2trash', side_effect=[None, None, ValueError])
    result = runner.invoke(mvid_rename_main, [str(d), str(non_dir), str(d)])
    assert result.exit_code == 0


def test_cddb_query_main(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_result = mocker.MagicMock(_asdict=lambda: {'foo': 'bar'})
    mocker.patch('deltona.commands.media.cddb_query', return_value=mock_result)
    result = runner.invoke(cddb_query_main, ['arg1', 'arg2'])
    assert result.exit_code == 0
    assert '"foo": "bar"' in result.output


def test_ke_ebook_ex_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    d = tmp_path / 'dir'
    d.mkdir()
    mocker.patch('deltona.commands.media.unpack_ebook')
    mocker.patch('deltona.commands.media.send2trash')
    result = runner.invoke(ke_ebook_ex_main, [str(d), '--delete-paths'])
    assert result.exit_code == 0


def test_ke_ebook_ex_main_no_delete(mocker: MockerFixture, runner: CliRunner,
                                    tmp_path: Path) -> None:
    d = tmp_path / 'dir'
    d.mkdir()
    mocker.patch('deltona.commands.media.unpack_ebook')
    mock_send2trash = mocker.patch('deltona.commands.media.send2trash')
    result = runner.invoke(ke_ebook_ex_main, [str(d)])
    assert result.exit_code == 0
    mock_send2trash.assert_not_called()


def test_encode_dashcam_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    f = tmp_path / 'front'
    r = tmp_path / 'rear'
    o = tmp_path / 'out'
    f.mkdir()
    r.mkdir()
    mocker.patch('deltona.commands.media.archive_dashcam_footage')
    result = runner.invoke(encode_dashcam_main, [str(f), str(r), str(o)])
    assert result.exit_code == 0


def test_encode_dashcam_main_same_dirs(runner: CliRunner, tmp_path: Path) -> None:
    d = tmp_path / 'dir'
    d.mkdir()
    result = runner.invoke(encode_dashcam_main, [str(d), str(d), str(tmp_path / 'out')])
    assert result.exit_code != 0


def test_hlg2sdr_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    f = tmp_path / 'file.mkv'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.hlg_to_sdr')
    result = runner.invoke(hlg2sdr_main, [str(f)])
    assert result.exit_code == 0


def test_tbc2srt_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    f = tmp_path / 'file.tbc'
    f.write_text('dummy')
    mocker.patch('deltona.commands.media.sp.run')
    mocker.patch('deltona.commands.media.send2trash')
    result = runner.invoke(tbc2srt_main, [str(f)])
    assert result.exit_code == 0


def test_flac_dir_finalize_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    d = tmp_path / 'album'
    d.mkdir()
    flac = d / '01.flac'
    flac.write_text('dummy')
    img = d / 'cover.jpg'
    img.write_text('dummy')
    mocker.patch('deltona.commands.media.underscorize', side_effect=lambda x: x)
    mocker.patch('deltona.commands.media.make_sfv')
    mocker.patch(
        'deltona.commands.media.sp.run',
        return_value=mocker.MagicMock(
            stdout='tracknumber=1\nartist=Artist\ntitle=Title\nunknown\n'),
    )
    result = runner.invoke(flac_dir_finalize_main, [str(d)])
    assert result.exit_code == 0



from __future__ import annotations

from typing import TYPE_CHECKING

from deltona.adp import SalaryResponse, calculate_salary
from requests.exceptions import HTTPError
import pytest

if TYPE_CHECKING:
    from requests_mock import Mocker


def test_salary_response_str() -> None:
    response = SalaryResponse(federal=100.0,
                              fica=50.0,
                              fuckery=175.0,
                              gross=1000.0,
                              medicare=25.0,
                              net_pay=825.0,
                              state=0)
    expected_str = """Gross     \033[1;32m 1000.00\033[0m
Federal   \033[1;32m  100.00\033[0m
FICA      \033[1;32m   50.00\033[0m
Medicare  \033[1;32m   25.00\033[0m
State     \033[1;32m    0.00\033[0m
------------------
Net       \033[1;32m  825.00\033[0m

------------------
Fuckery   \033[1;31m  175.00\033[0m"""
    assert str(response) == expected_str


FEDERAL_TAX = 100.0
FICA_TAX = 50.0
GROSS_PAY = 4900.0
MEDICARE_TAX = 25.0
NET_PAY = 825.0
STATE = 0
FUCKERY = 175.0


def test_calculate_salary_success(requests_mock: Mocker) -> None:
    mock_response = {
        'content': {
            'federal': 100.0,
            'fica': 50.0,
            'medicare': 25.0,
            'netPay': 825.0,
            'state': 0
        }
    }
    requests_mock.post('https://calculators.symmetry.com/api/calculators/hourly?report=none',
                       json=mock_response)
    result = calculate_salary(hours=70, pay_rate=70.0, state='FL')

    assert result.federal == FEDERAL_TAX
    assert result.fica == FICA_TAX
    assert result.medicare == MEDICARE_TAX
    assert result.net_pay == NET_PAY
    assert result.state == STATE
    assert result.gross == GROSS_PAY


def test_calculate_salary_http_error(requests_mock: Mocker) -> None:
    requests_mock.post('https://calculators.symmetry.com/api/calculators/hourly?report=none',
                       status_code=500)
    with pytest.raises(HTTPError):
        calculate_salary(hours=70, pay_rate=70.0, state='FL')



from __future__ import annotations

from typing import TYPE_CHECKING
import plistlib

from deltona.commands.string import (
    cssq_main,
    fullwidth2ascii_main,
    is_ascii_main,
    is_bin_main,
    json2yaml_main,
    pl2json_main,
    sanitize_main,
    slugify_main,
    title_fixer_main,
    trim_main,
    ucwords_main,
    underscorize_main,
    urldecode_main,
)
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


@pytest.mark.parametrize(
    ('input_str', 'expected'),
    [
        ('hello', 0),
        # cspell: disable-next-line  # noqa: ERA001
        ('héllo', 1),
    ],
)
def test_is_ascii_main(runner: CliRunner, mocker: MockerFixture, tmp_path: Path, input_str: str,
                       expected: int) -> None:
    input_file = tmp_path / 'test_is_ascii.txt'
    input_file.write_text(input_str, encoding='utf-8')
    mocker.patch('deltona.string.is_ascii', return_value=expected == 0)
    result = runner.invoke(is_ascii_main, [str(input_file)])
    assert result.exit_code == expected


def test_urldecode_main_basic(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_urldecode.txt'
    input_file.write_text('hello%20world%21\n', encoding='utf-8')
    result = runner.invoke(urldecode_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'hello world!' in result.output


def test_urldecode_main_netloc(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_urldecode_netloc.txt'
    # cspell: disable-next-line  # noqa: ERA001
    input_file.write_text('https%3A%2F%2Fexample.com%2Ffoo\n', encoding='utf-8')
    mocker.patch('sys.argv', ['netloc'])
    result = runner.invoke(urldecode_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'example.com' in result.output


def test_underscorize_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_underscorize.txt'
    input_file.write_text('Hello World\n', encoding='utf-8')
    result = runner.invoke(underscorize_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'Hello_World' in result.output


def test_json2yaml_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_json2yaml.json'
    input_file.write_text('{"a": 1, "b": 2}', encoding='utf-8')
    result = runner.invoke(json2yaml_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'a: 1' in result.output


def test_json2yaml_main_flow_style(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_json2yaml_flow.json'
    input_file.write_text('{"a": 1, "b": 2}', encoding='utf-8')
    result = runner.invoke(json2yaml_main, ['-d', str(input_file)])
    assert result.exit_code == 0
    assert '{a: 1}' in result.output or 'a: 1' in result.output


def test_sanitize_main_default(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_sanitize.txt'
    input_file.write_text('hello<>:"/\\|?*\n', encoding='utf-8')
    result = runner.invoke(sanitize_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'hello' in result.output


def test_sanitize_main_no_restricted(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_sanitize_no_restricted.txt'
    input_file.write_text('hello<>:"/\\|?*\n', encoding='utf-8')
    result = runner.invoke(sanitize_main, ['-R', str(input_file)])
    assert result.exit_code == 0
    assert 'hello' in result.output


def test_trim_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_trim.txt'
    input_file.write_text('  hello  \n  world  \n', encoding='utf-8')
    result = runner.invoke(trim_main, [str(input_file)])
    assert result.exit_code == 0
    assert result.output == 'hello\nworld\n'


def test_ucwords_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_ucwords.txt'
    input_file.write_text('hello world\n', encoding='utf-8')
    result = runner.invoke(ucwords_main, [str(input_file)])
    assert result.exit_code == 0
    assert result.output == 'Hello World\n'


def test_pl2json_main_valid(runner: CliRunner, tmp_path: Path) -> None:
    data = {'foo': 'bar'}
    pl_file = tmp_path / 'test.plist'
    with pl_file.open('wb') as f:
        plistlib.dump(data, f)
    result = runner.invoke(pl2json_main, [str(pl_file)])
    assert '"foo": "bar"' in result.output


def test_pl2json_main_invalid(runner: CliRunner, tmp_path: Path) -> None:
    pl_file = tmp_path / 'test.plist'
    pl_file.write_bytes(b'not a plist')
    result = runner.invoke(pl2json_main, [str(pl_file)])
    assert result.exit_code != 0


def test_pl2json_main_cannot_serialize(runner: CliRunner, tmp_path: Path,
                                       mocker: MockerFixture) -> None:
    mocker.patch('deltona.commands.string.plistlib.load')
    mocker.patch('deltona.commands.string.json.dumps', side_effect=TypeError)
    pl_file = tmp_path / 'test.plist'
    pl_file.touch()
    result = runner.invoke(pl2json_main, [str(pl_file)])
    assert result.exit_code != 0
    assert 'A non-JSON serialisable item is present in the file.' in result.output


def test_is_bin_main_binary(runner: CliRunner, tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch('deltona.commands.string.is_binary_string', return_value=True)
    file = tmp_path / 'binfile'
    file.write_bytes(b'\x00\x01\x02')
    result = runner.invoke(is_bin_main, [str(file)])
    assert result.exit_code == 0


def test_is_bin_main_not_binary(runner: CliRunner, tmp_path: Path) -> None:
    file = tmp_path / 'ascii_file'
    file.write_text('abc\n')
    result = runner.invoke(is_bin_main, [str(file)])
    assert result.exit_code == 1


def test_is_bin_main_empty_not_binary(runner: CliRunner, tmp_path: Path) -> None:
    file = tmp_path / 'ascii_file'
    file.write_text('')
    result = runner.invoke(is_bin_main, [str(file)])
    assert result.exit_code == 1


def test_fullwidth2ascii_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_fullwidth2ascii.txt'
    # cspell: disable-next-line  # noqa: ERA001
    input_file.write_text('ｈｅｌｌｏ\n', encoding='utf-8')  # noqa: RUF001
    result = runner.invoke(fullwidth2ascii_main, [str(input_file)])
    assert 'hello' in result.output


def test_slugify_main(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_slugify.txt'
    input_file.write_text('Hello, World!\n', encoding='utf-8')
    result = runner.invoke(slugify_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'hello-world' in result.output


def test_slugify_main_no_lower(runner: CliRunner, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_slugify_no_lower.txt'
    input_file.write_text('Hello, World!\n', encoding='utf-8')
    result = runner.invoke(slugify_main, ['--no-lower', str(input_file)])
    assert result.exit_code == 0
    assert 'Hello-World' in result.output


def test_title_fixer_main_basic(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_title_fixer.txt'
    input_file.write_text('some title\n', encoding='utf-8')
    mocker.patch('deltona.naming.adjust_title', return_value='Fixed Title')
    result = runner.invoke(title_fixer_main, [str(input_file)])
    assert result.exit_code == 0
    assert 'Fixed Title' in result.output


def test_title_fixer_main_no_modes(runner: CliRunner, tmp_path: Path) -> None:
    result = runner.invoke(title_fixer_main, ['--no-english'], input='some title\n')
    assert result.exit_code != 0
    assert 'No modes specified.' in result.output


def test_cssq_prints_json_lines(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_cssq.html'
    input_file.write_text('<div>First</div><div>Second</div>', encoding='utf-8')
    mocker.patch('deltona.commands.string.cssq', return_value=['A', 'B'])
    mock_json = mocker.patch(
        'deltona.commands.string.json.dumps',
        side_effect=lambda x, *a, **kw: f'"{x}"',  # noqa: ARG005
    )
    result = runner.invoke(cssq_main, ['-j', '-s', '-t', 'div', str(input_file)])
    assert result.exit_code == 0
    assert mock_json.call_count == 2
    assert result.output == '"A"\n"B"\n'


def test_cssq_prints_strings(runner: CliRunner, mocker: MockerFixture, tmp_path: Path) -> None:
    input_file = tmp_path / 'test_cssq.html'
    input_file.write_text('<div>First</div><div>Second</div>', encoding='utf-8')
    mocker.patch('deltona.commands.string.cssq', return_value=['A', 'B'])
    mock_json = mocker.patch('deltona.commands.string.json.dumps')
    result = runner.invoke(cssq_main, ['-s', '-t', 'div', str(input_file)])
    assert result.exit_code == 0
    assert mock_json.call_count == 0
    assert result.output == 'A\nB\n'



from __future__ import annotations

from typing import TYPE_CHECKING, Any
import errno

from deltona.commands.desktop import (
    connect_g603_main,
    inhibit_notifications_main,
    kill_gamescope_main,
    mpv_sbs_main,
    umpv_main,
    upload_to_imgbb_main,
)
from requests import HTTPError

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_inhibit_notifications_main_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_inhibit = mocker.patch('deltona.commands.desktop.inhibit_notifications')
    mock_inhibit.return_value = 0
    result = runner.invoke(inhibit_notifications_main, [])
    assert result.exit_code == 0
    mock_inhibit.assert_called_once()


def test_inhibit_notifications_main_failure(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_sleep = mocker.patch('deltona.commands.desktop.sleep')
    mock_inhibit = mocker.patch('deltona.commands.desktop.inhibit_notifications')
    mock_inhibit.return_value = False
    runner.invoke(inhibit_notifications_main, [])
    mock_sleep.assert_not_called()


def test_inhibit_notifications_main_args(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_sleep = mocker.patch('deltona.commands.desktop.sleep')
    mock_inhibit = mocker.patch('deltona.commands.desktop.inhibit_notifications')
    result = runner.invoke(inhibit_notifications_main, ['--sleep-time', '10'])
    assert result.exit_code == 0
    mock_inhibit.assert_called_once()
    mock_sleep.assert_called_once_with(10)


def test_umpv_main_success(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.user_state_path')
    mock_socket = mocker.patch('deltona.commands.desktop.socket.socket')
    mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file"1.mp4'
    out_mp4.write_bytes(b'')
    result = runner.invoke(umpv_main, [str(out_mp4)])
    assert result.exit_code == 0
    mock_socket.return_value.send.assert_called_once_with(
        f'raw loadfile "{tmp_path}/file\\"1.mp4"\n'.encode())


def test_umpv_main_socket_conn_refused(mocker: MockerFixture, runner: CliRunner,
                                       tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.user_state_path', return_value=tmp_path)
    mock_socket = mocker.patch('deltona.commands.desktop.socket.socket')
    sock_exc = OSError(errno.ECONNREFUSED, 'Connection refused')
    mock_socket.return_value.connect.side_effect = sock_exc
    sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file"1.mp4'
    out_mp4.write_bytes(b'')
    result = runner.invoke(umpv_main, [str(out_mp4)])
    assert result.exit_code == 0
    mock_socket.return_value.send.assert_not_called()
    sp_run.assert_called_once_with(
        (
            'mpv',
            '--no-terminal',
            '--force-window',
            f'--input-ipc-server={tmp_path}/umpv-socket',
            '--',
            str(out_mp4),
        ),
        check=True,
    )


def test_umpv_main_socket_enoent(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.user_state_path', return_value=tmp_path)
    mock_socket = mocker.patch('deltona.commands.desktop.socket.socket')
    sock_exc = OSError(errno.ENOENT, 'No such file or directory')
    mock_socket.return_value.connect.side_effect = sock_exc
    sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file"1.mp4'
    out_mp4.write_bytes(b'')
    result = runner.invoke(umpv_main, [str(out_mp4)])
    assert result.exit_code == 0
    mock_socket.return_value.send.assert_not_called()
    sp_run.assert_called_once_with(
        (
            'mpv',
            '--no-terminal',
            '--force-window',
            f'--input-ipc-server={tmp_path}/umpv-socket',
            '--',
            str(out_mp4),
        ),
        check=True,
    )


def test_umpv_main_unhandled_socket_exception(mocker: MockerFixture, runner: CliRunner,
                                              tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.user_state_path', return_value=tmp_path)
    mock_socket = mocker.patch('deltona.commands.desktop.socket.socket')
    sock_exc = OSError(errno.ENOTBLK, 'Block device required')
    mock_socket.return_value.connect.side_effect = sock_exc
    sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file"1.mp4'
    out_mp4.write_bytes(b'')
    result = runner.invoke(umpv_main, [str(out_mp4)])
    assert result.exit_code != 0
    mock_socket.return_value.send.assert_not_called()
    sp_run.assert_not_called()


def test_connect_g603_import_error(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_system_bus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_system_bus_callable.side_effect = ImportError("No module named 'pydbus'")
    result = runner.invoke(connect_g603_main)
    assert result.exit_code != 0


def test_connect_g603_not_linux(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.IS_LINUX', False)  # noqa: FBT003
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mocker.patch('deltona.commands.desktop._get_pydbus_system_bus_callable')
    result = runner.invoke(connect_g603_main)
    assert result.exit_code != 0


def test_connect_g603(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G603'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.return_value = mock_device
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.side_effect = [
            [
                'org.bluez.Adapter1',
                {
                    'Discovering': True
                },
            ],
            [
                'org.bluez.Device1',
                {
                    'Address': '00:11:22:33:44:55',
                    'ServicesResolved': True,
                    'Connected': True,
                    'RSSI': -50,
                },
            ],
        ]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_remove_device = mock_bus.get.return_value.RemoveDevice
    mock_start_discovery = mock_bus.get.return_value.StartDiscovery
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_remove_device.assert_called_once_with('path')
    mock_start_discovery.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.assert_called_once()
    mock_device1.Pair.assert_called_once()
    assert 'Pairing with 00:11:22:33:44:55' in result.output


def test_connect_g603_unhandled_property_change(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_log_debug = mocker.patch('deltona.commands.desktop.log.debug')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G603'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.return_value = mock_device
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.return_value = ['org.bluez.Device1', {}]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_remove_device = mock_bus.get.return_value.RemoveDevice
    mock_start_discovery = mock_bus.get.return_value.StartDiscovery
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_remove_device.assert_called_once_with('path')
    mock_start_discovery.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.assert_called_once()
    mock_log_debug.assert_has_calls([
        mocker.call(
            'Unhandled property changes: interface=%s, values=%s, mac=%s',
            'org.bluez.Device1',
            {},
            '00:11:22:33:44:55',
        )
    ])


def test_connect_g603_disconnected(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_log_debug = mocker.patch('deltona.commands.desktop.log.debug')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G603'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.return_value = mock_device
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.side_effect = [
            ['org.bluez.Device1', {
                'ServicesResolved': False
            }],
            ['other_interface', {}],
        ]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_remove_device = mock_bus.get.return_value.RemoveDevice
    mock_start_discovery = mock_bus.get.return_value.StartDiscovery
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_remove_device.assert_called_once_with('path')
    mock_start_discovery.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.assert_called_once()
    mock_log_debug.assert_has_calls(
        [mocker.call('Device %s was disconnected.', '00:11:22:33:44:55')])


def test_connect_g603_ignores_non_g603_devices(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_log_debug = mocker.patch('deltona.commands.desktop.log.debug')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G604'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.return_value = mock_device
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.return_value = ['org.bluez.Device1', {}]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_remove_device = mock_bus.get.return_value.RemoveDevice
    mock_start_discovery = mock_bus.get.return_value.StartDiscovery
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_remove_device.assert_called_once_with('path')
    mock_start_discovery.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.assert_called_once()
    mock_log_debug.assert_has_calls(
        [mocker.call('Ignoring device %s (MAC: %s).', 'G604', '00:11:22:33:44:55')])


def test_connect_g603_key_error(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_log_debug = mocker.patch('deltona.commands.desktop.log.debug')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G603'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.side_effect = [mock_device, KeyError]
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.return_value = ['org.bluez.Device1', {}]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.assert_called_once()
    mock_log_debug.assert_has_calls(
        [mocker.call('Caught error with device %s: %s', '00:11:22:33:44:55', mocker.ANY)])


def test_connect_g603_paired(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_log_debug = mocker.patch('deltona.commands.desktop.log.debug')
    mock_get_gi_repository_glib = mocker.patch('deltona.commands.desktop._get_gi_repository_glib')
    mock_get_pydbus_callable = mocker.patch(
        'deltona.commands.desktop._get_pydbus_system_bus_callable')
    mock_bus = mock_get_pydbus_callable.return_value.return_value
    mock_device = mocker.MagicMock()
    mock_device.Name = 'G603'
    mock_device1 = mocker.MagicMock()
    mock_device.__getitem__.return_value = mock_device1
    mock_bus.get.return_value = mock_device
    mock_signal_subscribe = mock_bus.con.signal_subscribe

    def do_on_properties_changed(
        sender: str | None,
        interface_name: str | None,
        member: str | None,
        object_path: str | None,
        arg0: str | None,
        flags: int,
        callback: Callable[..., object],
        *user_data: Any,
    ) -> int:
        props = mocker.MagicMock()
        props.unpack.return_value = ['org.bluez.Device1', {'Paired': True}]
        callback(None, None, '/org/bluez/hci0/dev_00_11_22_33_44_55', None, None, props)
        return 0

    mock_signal_subscribe.side_effect = do_on_properties_changed
    mocker.patch(
        'deltona.commands.desktop.find_bluetooth_device_info_by_name',
        side_effect=[('path', {
            'Address': '00:11:22:33:44:55'
        }), KeyError],
    )
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.side_effect = (
        KeyboardInterrupt)
    result = runner.invoke(connect_g603_main)
    assert result.exit_code == 0
    mock_get_gi_repository_glib.return_value.MainLoop.return_value.run.assert_called_once()
    assert mock_get_gi_repository_glib.return_value.MainLoop.return_value.quit.call_count == 2
    mock_log_debug.assert_has_calls([mocker.call('Quitting.')])


def test_kill_gamescope_main(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_kill_gamescope = mocker.patch('deltona.commands.desktop.kill_gamescope')
    result = runner.invoke(kill_gamescope_main)
    assert result.exit_code == 0
    mock_kill_gamescope.assert_called_once()


def test_upload_to_imgbb_xdg_install(mocker: MockerFixture, runner: CliRunner,
                                     tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mock_requests_get = mocker.patch('deltona.commands.desktop.requests.get')
    mock_requests_get.return_value.content = b'some image data'
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    result = runner.invoke(upload_to_imgbb_main, ['--xdg-install', str(tmp_path)])
    assert result.exit_code == 0
    mock_sp_run.assert_called_once_with(
        ('update-desktop-database', '-v', str(tmp_path / 'share/applications')),
        check=True,
        capture_output=True,
    )
    assert (tmp_path / 'share/applications').exists()
    assert (tmp_path / 'share/applications/upload-to-imgbb.desktop').exists()
    assert ('TryExec=upload-to-imgbb'
            in (tmp_path / 'share/applications/upload-to-imgbb.desktop').read_text())
    assert (tmp_path / 'share/icons/hicolor/300x300/apps/imgbb.png').exists()


def test_upload_to_imgbb_no_gui(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value=None)
    mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mock_open = mocker.patch('deltona.commands.desktop.webbrowser.open')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.upload_to_imgbb',
        return_value=mocker.MagicMock(json=mocker.MagicMock(
            return_value={'data': {
                'url': 'https://example.com/image.png'
            }})),
    )
    a_png = tmp_path / 'a.png'
    a_png.write_bytes(b'some image data')
    result = runner.invoke(upload_to_imgbb_main,
                           [str(a_png), '--no-gui', '--no-browser', '--no-clipboard'])
    assert result.exit_code == 0
    mock_open.assert_not_called()
    assert 'https://example.com/image.png' in result.output


def test_upload_to_imgbb_gui(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value='fake')
    mock_copy = mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mock_open = mocker.patch('deltona.commands.desktop.webbrowser.open')
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.upload_to_imgbb',
        return_value=mocker.MagicMock(json=mocker.MagicMock(
            return_value={'data': {
                'url': 'https://example.com/image.png'
            }})),
    )
    a_png = tmp_path / 'a.png'
    a_png.write_bytes(b'some image data')
    result = runner.invoke(upload_to_imgbb_main, [str(a_png)])
    assert result.exit_code == 0
    mock_open.assert_not_called()
    assert 'https://example.com/image.png' in result.output
    mock_sp_run.assert_called_once_with(
        ('fake', '--title', 'Successfully uploaded', '--msgbox', 'https://example.com/image.png'),
        check=False,
    )
    mock_copy.assert_called_once_with('https://example.com/image.png')


def test_upload_to_imgbb_http_error_gui(mocker: MockerFixture, runner: CliRunner,
                                        tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value='fake')
    mock_copy = mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mock_open = mocker.patch('deltona.commands.desktop.webbrowser.open')
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.upload_to_imgbb', side_effect=HTTPError)
    a_png = tmp_path / 'a.png'
    a_png.write_bytes(b'some image data')
    result = runner.invoke(upload_to_imgbb_main, [str(a_png)])
    assert result.exit_code != 0
    assert 'Failed to upload. Check API key!' in result.output
    mock_open.assert_not_called()
    mock_sp_run.assert_called_once_with(('fake', '--sorry', 'Failed to upload!'), check=False)
    mock_copy.assert_not_called()


def test_upload_to_imgbb_http_error(mocker: MockerFixture, runner: CliRunner,
                                    tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value='fake')
    mock_copy = mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mock_open = mocker.patch('deltona.commands.desktop.webbrowser.open')
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch('deltona.commands.desktop.upload_to_imgbb', side_effect=HTTPError)
    a_png = tmp_path / 'a.png'
    a_png.write_bytes(b'some image data')
    result = runner.invoke(upload_to_imgbb_main, [str(a_png), '--no-gui'])
    assert result.exit_code != 0
    assert 'Failed to upload. Check API key!' in result.output
    mock_open.assert_not_called()
    mock_sp_run.assert_not_called()
    mock_copy.assert_not_called()


def test_upload_to_imgbb_no_files(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value=None)
    mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mocker.patch('deltona.commands.desktop.webbrowser.open')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.upload_to_imgbb',
        return_value=mocker.MagicMock(json=mocker.MagicMock(
            return_value={'data': {
                'url': 'https://example.com/image.png'
            }})),
    )
    result = runner.invoke(upload_to_imgbb_main)
    assert result.exit_code == 0


def test_upload_to_imgbb(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.which', return_value=None)
    mocker.patch('deltona.commands.desktop.pyperclip.copy')
    mock_open = mocker.patch('deltona.commands.desktop.webbrowser.open')
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.upload_to_imgbb',
        return_value=mocker.MagicMock(json=mocker.MagicMock(
            return_value={'data': {
                'url': 'https://example.com/image.png'
            }})),
    )
    a_png = tmp_path / 'a.png'
    a_png.write_bytes(b'some image data')
    result = runner.invoke(upload_to_imgbb_main, [str(a_png)])
    assert result.exit_code == 0
    mock_open.assert_called_once_with('https://example.com/image.png')


def test_mpv_sbs_main(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.ffprobe',
        return_value={
            'streams': [
                {
                    'codec_type': 'other',
                    'disposition': {
                        'default': 0
                    }
                },
                {
                    'codec_type': 'other'
                },
                {
                    'codec_type': 'video',
                    'width': 1920,
                    'height': 1080,
                    'disposition': {
                        'default': 1
                    },
                },
            ]
        },
    )
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file1.mp4'
    out_mp4.write_bytes(b'file1')
    out_mp4_right = tmp_path / 'file1_right.mp4'
    out_mp4_right.write_bytes(b'file1_right')
    result = runner.invoke(mpv_sbs_main, [str(out_mp4), str(out_mp4_right)])
    assert result.exit_code == 0
    mock_sp_run.assert_called_once_with(
        (
            'mpv',
            '--hwdec=no',
            '--config=no',
            str(out_mp4),
            f'--external-file={out_mp4_right}',
            '--lavfi-complex=[vid1][vid2] hstack [vo]',
        ),
        check=True,
    )


def test_mpv_sbs_main_non_matching_height(mocker: MockerFixture, runner: CliRunner,
                                          tmp_path: Path) -> None:
    mocker.patch('deltona.commands.desktop.setup_logging')
    mocker.patch(
        'deltona.commands.desktop.ffprobe',
        side_effect=[
            {
                'streams': [
                    {
                        'codec_type': 'other',
                        'disposition': {
                            'default': 0
                        }
                    },
                    {
                        'codec_type': 'other'
                    },
                    {
                        'codec_type': 'video',
                        'width': 1920,
                        'height': 1080,
                        'disposition': {
                            'default': 1
                        },
                    },
                ]
            },
            {
                'streams': [{
                    'codec_type': 'video',
                    'width': 640,
                    'height': 480,
                    'disposition': {
                        'default': 1
                    },
                }]
            },
        ],
    )
    mock_sp_run = mocker.patch('deltona.commands.desktop.sp.run')
    out_mp4 = tmp_path / 'file1.mp4'
    out_mp4.write_bytes(b'file1')
    out_mp4_right = tmp_path / 'file1_right.mp4'
    out_mp4_right.write_bytes(b'file1_right')
    result = runner.invoke(mpv_sbs_main, [str(out_mp4), str(out_mp4_right)])
    assert result.exit_code == 0
    mock_sp_run.assert_called_once_with(
        (
            'mpv',
            '--hwdec=no',
            '--config=no',
            str(out_mp4),
            f'--external-file={out_mp4_right}',
            ('--lavfi-complex=[vid1] scale=1920x1080 [vid1_scale];[vid2] scale=1920x1080 '
             '[vid2_crop];[vid1_scale][vid2_crop] hstack [vo]'),
        ),
        check=True,
    )



"""System utilities."""

from __future__ import annotations

from pathlib import Path
from shlex import quote
from time import sleep
from typing import TYPE_CHECKING, Any, cast
import configparser
import json
import logging
import os
import plistlib
import re
import subprocess as sp
import sys

from binaryornot.helpers import is_binary_string
from typing_extensions import override

from .io import context_os_open
from .string import slugify
from .typing import CDStatus, StrPath, StrPathMustExist

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pydbus.bus import OrgBluezDict as OrgBluezDevice1Dict

__all__ = (
    'CHROME_DEFAULT_CONFIG_PATH',
    'CHROME_DEFAULT_LOCAL_STATE_PATH',
    'IS_LINUX',
    'MultipleKeySlots',
    'find_bluetooth_device_info_by_name',
    'get_kwriteconfig_commands',
    'inhibit_notifications',
    'kill_gamescope',
    'kill_wine',
    'pan_connect',
    'pan_disconnect',
    'patch_macos_bundle_info_plist',
    'reset_tpm_enrollment',
    'slug_rename',
    'uninhibit_notifications',
    'wait_for_disc',
)

CDROM_DRIVE_STATUS = 0x5326
IS_LINUX = sys.platform == 'linux'
"""
If the system is Linux.

:meta hide-value:
"""
IS_WINDOWS = sys.platform == 'win32' or sys.platform == 'cygwin'
CHROME_DEFAULT_CONFIG_PATH = Path('~/.config/google-chrome').expanduser()
"""
Default Chrome config path.

:meta hide-value:
"""
CHROME_DEFAULT_LOCAL_STATE_PATH = str(CHROME_DEFAULT_CONFIG_PATH / 'Local State')
"""
Default Chrome local state path.

:meta hide-value:
"""
log = logging.getLogger(__name__)


def wait_for_disc(drive_path: StrPathMustExist = 'dev/sr0', *, sleep_time: float = 1.0) -> bool:
    """
    Wait for a disc to be ready in a drive.

    For Linux only.

    Parameters
    ----------
    drive_path : StrPathMustExist
        Path to the optical drive device.
    sleep_time : float
        Seconds to sleep between polling attempts.

    Returns
    -------
    bool
        ``True`` if the disc is ready, ``False`` if interrupted.
    """
    import fcntl  # noqa: PLC0415

    with context_os_open(drive_path, os.O_RDONLY | os.O_NONBLOCK) as f:
        s = -1
        try:
            while (s := fcntl.ioctl(f, CDROM_DRIVE_STATUS)) != CDStatus.DISC_OK:
                sleep(sleep_time)
        except KeyboardInterrupt:
            pass
    return s == CDStatus.DISC_OK


_key: int | None = None


def inhibit_notifications(name: str = __name__, reason: str = 'No reason specified.') -> bool:
    """
    Disable notifications.

    This function is meant to be called from a separate thread from the working task thread during
    initialisation.

    Parameters
    ----------
    name : str
        Name of application.
    reason : str
        Reason for inhibition.

    Returns
    -------
    bool
        ``True`` if inhibited, ``False`` otherwise.
    """
    global _key  # noqa: PLW0603
    try:
        from pydbus import SessionBus  # noqa: PLC0415
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        log.exception('Cannot import pydbus.', stack_info=False)
        return False
    notifications = SessionBus().get('org.freedesktop.Notifications',
                                     '/org/freedesktop/Notifications')
    if notifications.Inhibited:
        return False
    log.debug('Disabling notifications.')
    _key = notifications.Inhibit(name, reason, {})
    return True


def uninhibit_notifications() -> None:
    """
    Enable notifications.

    This function is meant to be called from a separate thread from the working task thread during
    clean-up.

    This function is not required to be called if the application will exit as the state will be
    restored on clean-up of the D-Bus connection.

    Raises
    ------
    ConnectionError
        If D-Bus connection is not available (which can be caused by not having pydbus installed).
    """
    global _key  # noqa: PLW0603
    try:
        from pydbus import SessionBus  # noqa: PLC0415
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        log.exception('Cannot import pydbus.', stack_info=False)
        return
    notifications = SessionBus().get('org.freedesktop.Notifications',
                                     '/org/freedesktop/Notifications')
    if not notifications:
        raise ConnectionError
    if not notifications.Inhibited:
        return
    if _key is not None:
        notifications.UnInhibit(_key)
        _key = None


def get_inhibitor(what: str, who: str, why: str, mode: str) -> int:
    try:
        from pydbus import SystemBus  # noqa: PLC0415
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        log.exception('Cannot import pydbus.', stack_info=False)
        return -1
    login1 = SystemBus().get('org.freedesktop.login1', '/org/freedesktop/login1')
    return cast('int', login1['org.freedesktop.login1.Manager'].Inhibit(what, who, why, mode))


def find_bluetooth_device_info_by_name(name: str) -> tuple[str, OrgBluezDevice1Dict]:
    """
    Get Bluetooth device information from D-Bus (bluez) by name.

    Note that not all devices present a name.

    Parameters
    ----------
    name : str
        Name of the device. Case sensitive.

    Returns
    -------
    tuple[str, OrgBluezDevice1Dict]
        Returns the D-Bus object path and a dictionary representing the ``org.bluez.Device1``
        properties.

    Raises
    ------
    NotImplementedError
        If not on Linux.
    KeyError
        If no device is found.
    """
    if not IS_LINUX:
        raise NotImplementedError
    from pydbus import SystemBus  # noqa: PLC0415

    bluez = SystemBus().get('org.bluez', '/')
    for k, v in bluez['org.freedesktop.DBus.ObjectManager'].GetManagedObjects().items():
        if ('org.bluez.Device1' in v and 'Name' in v['org.bluez.Device1']
                and v['org.bluez.Device1']['Name'] == name):
            return k, v['org.bluez.Device1']
    raise KeyError(name)


def pan_connect(device_mac: str, hci: str = 'hci0') -> None:
    """
    Connect to a Bluetooth PAN (Personal Area Network) device.

    Parameters
    ----------
    device_mac : str
        MAC address of the Bluetooth device.
    hci : str
        HCI adapter name.

    Raises
    ------
    NotImplementedError
        If not on Linux.
    """
    if not IS_LINUX:
        raise NotImplementedError
    from pydbus import SystemBus  # noqa: PLC0415

    device_mac = f'dev_{device_mac.upper().replace(":", "_")}'
    device = SystemBus().get('org.bluez', f'/org/bluez/{hci}/{device_mac}')
    device.Connect('nap')


def pan_disconnect(device_mac: str, hci: str = 'hci0') -> None:
    """
    Disconnect from a Bluetooth PAN (Personal Area Network) device.

    Parameters
    ----------
    device_mac : str
        MAC address of the Bluetooth device.
    hci : str
        HCI adapter name.

    Raises
    ------
    NotImplementedError
        If not on Linux.
    """
    if not IS_LINUX:
        raise NotImplementedError
    from pydbus import SystemBus  # noqa: PLC0415

    device_mac = f'dev_{device_mac.upper().replace(":", "_")}'
    device = SystemBus().get('org.bluez', f'/org/bluez/{hci}/{device_mac}')
    device.Disconnect()


def slug_rename(path: StrPath, *, no_lower: bool = False) -> StrPath:
    """
    Perform a slug rename on a file or directory.

    Parameters
    ----------
    path : StrPath
        Path to the file or directory.
    no_lower : bool
        If ``True``, do not convert to lowercase.

    Returns
    -------
    StrPath
        The new path after renaming.
    """
    path = Path(path).resolve(strict=True)
    parent = path.parent
    return path.rename(parent / slugify(path.name, no_lower=no_lower))


def patch_macos_bundle_info_plist(bundle: StrPath, **data: Any) -> None:
    """
    Patch a macOS application bundle's ``Info.plist`` with additional data.

    Parameters
    ----------
    bundle : StrPath
        Path to the ``.app`` bundle.
    **data : Any
        Key-value pairs to merge into the plist.
    """
    info_plist = Path(bundle).resolve(strict=True) / 'Contents' / 'Info.plist'
    with info_plist.open('rb') as f:
        file_data: dict[str, Any] = plistlib.load(f)
    with info_plist.open('wb') as f:
        plistlib.dump(file_data | data, f, sort_keys=False)
    info_plist.touch()


def kill_gamescope() -> None:
    """Kill all running Gamescope and gamescopereaper processes."""
    import psutil  # noqa: PLC0415

    for proc in (x for x in psutil.process_iter(('pid', 'name', 'username'))
                 if x.info['name'] in {'gamescope', 'gamescopereaper'}):
        proc.kill()


def kill_wine() -> None:
    """Kill all running Wine processes including wineserver and Wine-wrapped executables."""
    import psutil  # noqa: PLC0415

    for proc in (x for x in psutil.process_iter(('pid', 'name', 'username'))
                 if x.info['name'] in {'wineserver', 'wine-preloader', 'wine64-preloader'} or (
                     x.info['name'].lower().endswith('.exe'))):
        proc.kill()


class MultipleKeySlots(Exception):
    """Raised when a LUKS device has more than one keyslot."""
    @override
    def __init__(self, dev: str) -> None:
        super().__init__(f'Device {dev} has more than one keyslot. This is not supported.')


def reset_tpm_enrollment(uuid: str, *, dry_run: bool = True) -> None:
    """
    Reset TPM2 enrolment for a LUKS-encrypted device.

    Parameters
    ----------
    uuid : str
        UUID of the LUKS device.
    dry_run : bool
        If ``True``, only log the commands that would be run.

    Raises
    ------
    MultipleKeySlots
        If the device has more than one keyslot associated with the TPM2 token.
    """
    dev = f'/dev/disk/by-uuid/{uuid}'
    cmd = ('cryptsetup', 'luksDump', '--dump-json-metadata', dev)
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    info = json.loads(sp.run(cmd, check=True, capture_output=True).stdout)
    systemd_tokens = [(k, x) for k, x in info['tokens'].items() if x['type'] == 'systemd-tpm2']
    if not systemd_tokens:
        log.debug('No tokens found for device %s.', dev)
        return
    systemd_token_id = systemd_tokens[0][0]
    n_keyslots = len(info['tokens'][systemd_token_id]['keyslots'])
    if n_keyslots > 1:
        raise MultipleKeySlots(dev)
    slot = info['tokens'][systemd_token_id]['keyslots'][0]
    luks_kill_slot_cmd = ('cryptsetup', 'luksKillSlot', dev, slot)
    luks_kill_slot_cmd_quoted = ' '.join(quote(x) for x in luks_kill_slot_cmd)
    luks_token_remove_cmd = ('cryptsetup', 'token', 'remove', f'--token-id={systemd_token_id}', dev)
    luks_token_remove_cmd_quoted = ' '.join(quote(x) for x in luks_token_remove_cmd)
    cryptenroll_cmd = ('systemd-cryptenroll', '--tpm2-device=auto', dev)
    cryptenroll_cmd_quoted = ' '.join(quote(x) for x in cryptenroll_cmd)
    if dry_run:
        log.info('Would run: %s', luks_kill_slot_cmd_quoted)
        log.info('Would run: %s', luks_token_remove_cmd_quoted)
        log.info('Would run: %s', cryptenroll_cmd_quoted)
    else:
        log.debug('Running: %s', luks_kill_slot_cmd_quoted)
        sp.run(luks_kill_slot_cmd, check=True)
        log.debug('Running: %s', luks_token_remove_cmd_quoted)
        sp.run(luks_token_remove_cmd, check=True)
        log.debug('Running: %s', cryptenroll_cmd_quoted)
        sp.run(cryptenroll_cmd, check=True)
        log.info('Reset TPM enrolment for %s.', uuid)


IGNORED_GROUPS = {
    'KFileDialog Settings',
    'FileDialogSize',
    'Recent Files[$e]',
    'Recent URLs[$e]',
    'Recent Files',
    '$Version',
}
DEFAULT_FILE = Path.home() / '.config' / 'kdeglobals'
POSITION_RE = (
    r'(^(Height|Width|Window-Maximized) [0-9]+)|'
    r'((e?DP-[0-9]+|HDMI-[0-9]+(-[0-9]+)?|VNC-[0-9]+)$)|'
    r'((e?DP-[0-9]+|HDMI-[0-9]+(-[0-9]+)?|VNC-[0-9]+) (Height|Width|(X|Y)Position|Window-Maximized))|'  # noqa: E501
    r'([0-9]+x[0-9]+ screen: (Height|Width|(X|Y)Position)$)|'
    r'([0-9] screens: (Height|Width|(X|Y)Position)$)')
STATE_RE = r'^AAAA/'


def _iter_config_sections(file: StrPath) -> Iterator[tuple[str, str, str]]:
    config = configparser.ConfigParser(delimiters=('=',), interpolation=None)
    config.optionxform = str  # type: ignore[assignment]
    try:
        config.read(file)
    except (UnicodeDecodeError, configparser.MissingSectionHeaderError, configparser.ParsingError):
        log.warning('Cannot parse "%s".', file)
        return
    for section in config.sections():
        if '][' in section:
            log.debug('Skipping unsupported section "%s".', section)
            continue
        if section in IGNORED_GROUPS:
            log.debug('Ignoring section "%s".', section)
            continue
        for key, value in config[section].items():
            if is_binary_string(value.encode()):
                log.debug('Ignoring binary value in key %s.', key)
                continue
            if re.search(POSITION_RE, key):
                log.debug('Skipping metrics key "%s".', key)
                continue
            if key == 'State' and re.search(STATE_RE, value):
                log.debug('Skipping state key.')
                continue
            if key.endswith('[$e]'):
                log.debug('Skipping special key "%s".', key)
                continue
            yield section, key, value


def get_kwriteconfig_commands(file: StrPath = DEFAULT_FILE) -> Iterator[str]:
    """
    Generate ``kwriteconfig6`` commands from a KDE configuration file.

    Parameters
    ----------
    file : StrPath
        Path to the KDE configuration file.

    Yields
    ------
    str
        A ``kwriteconfig6`` command string.
    """
    home = str(Path.home())
    file = Path(file).resolve(strict=True)
    is_default_file = file == DEFAULT_FILE
    displayed_file = re.sub(rf'^{home}/', '~/', str(file))
    for section, key, value in _iter_config_sections(file):
        type_ = None
        re.match(r'^-?[0-9]+$', value)
        try:
            is_file = '/' in value and Path(value).exists()
        except OSError:
            is_file = False
        if is_file:
            type_ = 'path'
        elif re.match(r'^(?:1|true|false|on|yes)$', value, flags=re.IGNORECASE):
            type_ = 'bool'
        elif re.match(r'^-?[0-9]+$', value):
            type_ = 'int'
        cmd: tuple[str, ...] = ('kwriteconfig6',)
        if not is_default_file:
            cmd += ('--file', displayed_file)
        if type_:
            cmd += ('--type', quote(type_))
        cmd += ('--group', quote(section), '--key', quote(key), quote(value))
        yield ' '.join(cmd)


def get_kconfig_dict(file: StrPath = DEFAULT_FILE) -> dict[str, dict[str, str | int | bool]]:
    file = Path(file).resolve(strict=True)
    ret: dict[str, dict[str, str | int | bool]] = {}
    for section, key, value_ in _iter_config_sections(file):
        value: str | int | bool
        if re.match(r'^(?:1|true|false|on|yes)$', value_, flags=re.IGNORECASE):
            value = bool(re.match(r'^(?:1|true|on|yes)', value_, flags=re.IGNORECASE))
        elif re.match(r'^-?[0-9]+$', value_):
            value = int(value_)
        else:
            value = value_
        try:
            ret[section][key] = value
        except KeyError:
            ret[section] = {key: value}
    return ret



"""Salary calculator."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import TYPE_CHECKING, TypedDict, cast

from typing_extensions import override
import requests

from .string import strip_ansi_if_no_colors

if TYPE_CHECKING:
    from .typing import INCITS38Code

__all__ = ('SalaryResponse', 'calculate_salary')

# Find the API key by looking at requests on this page
# http://www.symmetry.com/try-it-for-free/calculators
API_KEY = 'RnFqNFA0NVlRTExEenRwWjNiRnJrTXY4WkZHZEpkcENEeFFzQ3F0Nnh5VT0='
POST_URI = 'https://calculators.symmetry.com/api/calculators/hourly?report=none'
REFERER = 'https://www.symmetry.com/'


class ContentDict(TypedDict):
    federal: float
    fica: float
    medicare: float
    netPay: float
    state: float


class ResponseDict(TypedDict):
    content: ContentDict


@dataclass
class SalaryResponse:
    """Response from the Symmetry API."""

    federal: float
    """Federal tax amount."""
    fica: float
    """FICA tax amount."""
    fuckery: float
    """Difference between gross and net pay."""
    gross: float
    """Gross pay amount."""
    medicare: float
    """Medicare tax amount."""
    net_pay: float
    """Net pay amount."""
    state: float
    """State tax amount."""
    @override
    def __str__(self) -> str:
        return strip_ansi_if_no_colors(f"""Gross     \033[1;32m{self.gross:8.2f}\033[0m
Federal   \033[1;32m{self.federal:8.2f}\033[0m
FICA      \033[1;32m{self.fica:8.2f}\033[0m
Medicare  \033[1;32m{self.medicare:8.2f}\033[0m
State     \033[1;32m{self.state:8.2f}\033[0m
------------------
Net       \033[1;32m{self.net_pay:8.2f}\033[0m

------------------
Fuckery   \033[1;31m{self.fuckery:8.2f}\033[0m""")


def calculate_salary(*,
                     hours: int = 70,
                     pay_rate: float = 70.0,
                     state: INCITS38Code = 'FL') -> SalaryResponse:
    """
    Calculate a US salary using the Symmetry API.

    Parameters
    ----------
    hours : int
        The number of hours worked. Default is 70.
    pay_rate : float
        The pay rate per hour. Default is 70.0.
    state : INCITS38Code
        The state code. Default is 'FL'.

    Returns
    -------
    SalaryResponse
        The response from the Symmetry API.
    """
    check_date = int(datetime.now(tz=timezone.utc).timestamp() * 1000)
    gross_pay = hours * pay_rate
    req = requests.post(
        POST_URI,
        headers={
            'accept': 'application/json, text/javascript, */*; q=0.01',
            'cache-control': 'no-cache',
            'dnt': '1',
            'origin': 'https://www.symmetry.com',
            'pcc-api-key': API_KEY,
            'referer': REFERER,
        },
        json={
            'checkDate': check_date,
            'state': state.upper(),
            'rates': [{
                'payRate': str(pay_rate),
                'hours': str(hours),
            }],
            'grossPay': str(gross_pay),
            'grossPayType': 'PAY_PER_PERIOD',
            'grossPayYTD': '0',
            'payFrequency': 'MONTHLY',
            'exemptFederal': 'false',
            'exemptFica': 'false',
            'exemptMedicare': 'false',
            'federalFilingStatusType': 'SINGLE',
            'federalAllowances': '0',
            'additionalFederalWithholding': '0',
            'roundFederalWithholding': 'false',
            'print': {
                'checkDate': check_date,
                'checkNumber': '',
                'checkNumberOnCheck': 'false',
                'companyAddressLine1': '',
                'companyAddressLine2': '',
                'companyAddressLine3': '',
                'companyName': '',
                'companyNameOnCheck': 'false',
                'employeeAddressLine1': '',
                'employeeAddressLine2': '',
                'employeeAddressLine3': '',
                'employeeName': '',
                'id': '',
                'remarks': '',
            },
            'otherIncome': [],
            'payCodes': [],
            'stockOptions': [],
            'stateInfo': {
                'parms': [
                    {
                        'name': 'TOTALALLOWANCES',
                        'value': '0'
                    },
                    {
                        'name': 'additionalStateWithholding',
                        'value': '0'
                    },
                    {
                        'name': 'SPOUSEBLINDNESS',
                        'value': 'false'
                    },
                    {
                        'name': 'stateExemption',
                        'value': 'false'
                    },
                    {
                        'name': 'PERSONALBLINDNESS',
                        'value': 'false'
                    },
                    {
                        'name': 'HEADOFHOUSEHOLD',
                        'value': 'false'
                    },
                    {
                        'name': 'FULLTIMESTUDENT',
                        'value': 'false'
                    },
                ]
            },
            'voluntaryDeductions': [],
            'presetDeductions': [],
        },
        timeout=30,
    )
    req.raise_for_status()
    data = cast('ResponseDict', req.json())['content']
    return SalaryResponse(
        federal=data['federal'],
        fica=data['fica'],
        fuckery=gross_pay - data['netPay'],
        gross=gross_pay,
        medicare=data['medicare'],
        net_pay=data['netPay'],
        state=data['state'],
    )



"""String commands."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TextIO
from urllib.parse import unquote_plus, urlparse
import io
import json
import plistlib
import sys

from bascom import setup_logging
from binaryornot.helpers import is_binary_string
from deltona import naming
from deltona.constants import CONTEXT_SETTINGS
from deltona.string import (
    cssq,
    fullwidth_to_narrow,
    is_ascii,
    sanitize,
    slugify,
    underscorize,
)
from deltona.typing import DecodeErrorsOption
import click
import yaml

if TYPE_CHECKING:
    from io import BytesIO


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def is_ascii_main(file: TextIO) -> None:
    """Check if a file is ASCII."""  # noqa: DOC501
    if not is_ascii(file.read()):
        raise click.exceptions.Exit(1)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-e', '--encoding', default='utf-8', help='Text encoding.')
@click.option(
    '-r',
    '--errors',
    default='strict',
    type=click.Choice(DecodeErrorsOption.__args__),  # type: ignore[attr-defined]
    help='Error handling mode.',
)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def urldecode_main(file: TextIO,
                   encoding: str = 'utf-8',
                   errors: DecodeErrorsOption = 'strict') -> None:
    """Decode a URL-encoded string."""
    ctx = click.get_current_context()
    is_netloc = ctx.info_name == 'netloc' or Path(sys.argv[0]).stem == 'netloc'
    for line in file:
        val = unquote_plus(line, encoding, errors)
        if is_netloc:
            val = urlparse(val).netloc.strip()
        click.echo(val.strip())


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def underscorize_main(file: TextIO) -> None:
    """Convert a string to an underscorised form."""
    for line in file:
        click.echo(underscorize(line.strip()))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
@click.option('-d', '--default-flow-style', is_flag=True, help='Enable compact flow style.')
@click.option('-i', '--indent', default=2, type=click.IntRange(2, 9), help='Indent width (spaces).')
def json2yaml_main(file: TextIO, indent: int = 2, *, default_flow_style: bool = False) -> None:
    """Convert JSON to YAML."""
    click.echo(yaml.dump(json.load(file), indent=indent, default_flow_style=default_flow_style))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
@click.option('-R', '--no-restricted', is_flag=True, help='Do not use restricted character set.')
def sanitize_main(file: TextIO, *, no_restricted: bool = False) -> None:
    """
    Transform a string to a 'sanitised' form.

    By default, a restricted character set safe for Windows filenames is used. Disable with -R.
    """
    click.echo(sanitize(file.read(), restricted=not no_restricted))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def trim_main(file: TextIO) -> None:
    """Trim lines in file."""
    for line in file:
        click.echo(line.strip())


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def ucwords_main(file: TextIO) -> None:
    """
    Run Python ``str.title()`` for lines in file.

    Named after PHP's function.
    """
    for line in file:
        click.echo(line.title(), nl=False)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('rb'), default=sys.stdin)
def pl2json_main(file: BytesIO) -> None:
    """
    Convert a Property List file to JSON.

    This command does not do any type conversions. This means files containing <data> objects will
    not work.
    """  # noqa: DOC501
    try:
        click.echo(json.dumps(plistlib.load(file), sort_keys=True, allow_nan=False, indent=2))
    except TypeError as e:
        click.echo('A non-JSON serialisable item is present in the file.', err=True)
        raise click.Abort from e


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('rb'), default=sys.stdin)
def is_bin_main(file: BytesIO) -> None:
    """
    Check if a file has binary contents.

    For this utility, 0 byte files do not count as binary.

    Exit code 0 means the file probably contains binary content.
    """  # noqa: DOC501
    file.seek(0, io.SEEK_END)
    if file.tell() == 0:
        click.echo('File is empty. Not counting as binary.', err=True)
        raise click.Abort
    file.seek(0)
    if is_binary_string(file.read(1024)):
        return
    raise click.exceptions.Exit(1)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
def fullwidth2ascii_main(file: TextIO) -> None:
    """Convert fullwidth characters to ASCII."""
    click.echo(fullwidth_to_narrow(file.read()), nl=False)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('file', type=click.File('r'), default=sys.stdin)
@click.option('--no-lower', is_flag=True, help='Disable lowercase.')
def slugify_main(file: TextIO, *, no_lower: bool = False) -> None:
    """Slugify a string."""
    click.echo(slugify(file.read(), no_lower=no_lower))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('titles', type=click.File('r'), default=sys.stdin)
@click.option('--no-names', help='Disable name checking.', is_flag=True)
@click.option('-E', '--no-english', help='Disable English mode.', is_flag=True)
@click.option('-a', '--arabic', help='Enable Arabic mode.', is_flag=True)
@click.option('-c', '--chinese', help='Enable Chinese mode.', is_flag=True)
@click.option('-j', '--japanese', help='Enable Japanese mode.', is_flag=True)
@click.option('-s', '--ampersands', help='Replace " and " with " & ".', is_flag=True)
def title_fixer_main(
    titles: tuple[str, ...],
    *,
    no_english: bool = False,
    chinese: bool = False,
    japanese: bool = False,
    arabic: bool = False,
    no_names: bool = False,
    ampersands: bool = False,
) -> None:
    """Fix titles."""  # noqa: DOC501
    modes = (
        *((naming.Mode.Arabic,) if arabic else ()),
        *((naming.Mode.Chinese,) if chinese else ()),
        *((naming.Mode.English,) if not no_english else ()),
        *((naming.Mode.Japanese,) if japanese else ()),
    )
    if not modes:
        click.echo('No modes specified.', err=True)
        raise click.Abort
    for title in titles:
        click.echo(naming.adjust_title(title, modes, disable_names=no_names, ampersands=ampersands))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('selector', nargs=1)
@click.argument('file', type=click.File('r'), default=sys.stdin)
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
@click.option('--debug-selector', is_flag=True, help='Enable SoupSieve debug logging.')
@click.option('-l',
              '--limit',
              default=0,
              type=click.IntRange(0, None),
              help='Limit number of results.')
@click.option('-j', '--json', 'json_lines', is_flag=True, help='Output as JSON lines.')
@click.option('-s', '--strip', is_flag=True, help='Strip whitespace from text content.')
@click.option('-t', '--text', is_flag=True, help='Output content within the HTML tags only.')
def cssq_main(
    file: TextIO,
    selector: str,
    limit: int = 0,
    *,
    debug: bool = False,
    debug_selector: bool = False,
    json_lines: bool = False,
    strip: bool = False,
    text: bool = False,
) -> None:
    """Filter HTML with CSS."""
    setup_logging(debug=debug, loggers={'bs4': {}, 'deltona': {}, 'soupsieve': {}})
    for item in cssq(selector,
                     file,
                     debug_selector=debug_selector,
                     limit=limit,
                     strip=strip,
                     text=text):
        click.echo(json.dumps(str(item)) if json_lines else str(item))



"""Uncategorised utilities."""

from __future__ import annotations

from http import HTTPStatus
from math import trunc
from pathlib import Path
from shlex import quote
from shutil import rmtree
from signal import SIGTERM
from typing import TYPE_CHECKING, overload
import csv
import logging
import os
import re
import subprocess as sp
import time

from requests.adapters import BaseAdapter
from typing_extensions import override
import requests

from .media import CD_FRAMES
from .system import IS_WINDOWS, kill_wine

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from paramiko import SFTPClient, SSHClient

    from .typing import StrPath

__all__ = (
    'DataAdapter',
    'add_cdda_times',
    'kill_processes_by_name',
    'secure_move_path',
    'unregister_wine_file_associations',
)

ZERO_TO_99 = '|'.join(f'{x:02d}' for x in range(100))
ZERO_TO_59 = '|'.join(f'{x:02d}' for x in range(60))
ZERO_TO_74 = '|'.join(f'{x:02d}' for x in range(75))
TIMES_RE = re.compile(f'^({ZERO_TO_99}):({ZERO_TO_59}):({ZERO_TO_74})$')
MAX_MINUTES = 99
MAX_SECONDS = 60
log = logging.getLogger(__name__)


def add_cdda_times(times: Iterable[str] | None) -> str | None:
    """
    Add CDDA time strings and get a total runtime in CDDA format.

    CDDA format is ``MM:SS:FF`` where ``MM`` is minutes, ``SS`` is seconds, and ``FF`` is frames.

    Parameters
    ----------
    times : Iterable[str] | None
        CDDA time strings to add together.

    Returns
    -------
    str | None
        The total runtime in CDDA format, or ``None`` if the input is empty or invalid.
    """
    if not times:
        return None
    total_ms = 0
    for time_ in times:
        if not (res := re.match(TIMES_RE, time_)):
            return None
        minutes, seconds, frames = [int(x) for x in res.groups()]
        total_ms += ((minutes * MAX_SECONDS * 1000) + (seconds * 1000) +
                     ((frames * 1000) // CD_FRAMES))
    minutes = total_ms // (MAX_SECONDS * 1000)
    remainder_ms = total_ms % (MAX_SECONDS * 1000)
    seconds = remainder_ms // 1000
    remainder_ms %= 1000
    frames = round((remainder_ms * 1000 * CD_FRAMES) / 1000000)
    if minutes > MAX_MINUTES or seconds > (MAX_SECONDS - 1) or frames > (CD_FRAMES - 1):
        return None
    return f'{trunc(minutes):02d}:{trunc(seconds):02d}:{trunc(frames):02d}'


def unregister_wine_file_associations(*, debug: bool = False) -> None:
    """
    Remove Wine file associations, icons, and MIME types from the user's desktop environment.

    Kills running Wine processes before removing files.

    Parameters
    ----------
    debug : bool
        If ``True``, pass verbose flags to ``update-desktop-database`` and ``update-mime-database``.
    """
    kill_wine()
    for item in (Path.home() / '.local/share/applications').glob('wine-extension-*.desktop'):
        log.debug('Removing file association "%s".', item)
        item.unlink()
    for item in (Path.home() / '.local/share/icons/hicolor').rglob('application-x-wine-extension*'):
        log.debug('Removing icon "%s".', item)
        item.unlink()
    (Path.home() / '.local/share/applications/mimeinfo.cache').unlink(missing_ok=True)
    for item in (Path.home() / '.local/share/mime/packages').glob('x-wine*'):
        log.debug('Removing MIME file "%s".', item)
        item.unlink()
    for item in (Path.home() / '.local/share/application').glob('x-wine-extension*'):
        log.debug('Removing MIME file "%s".', item)
        item.unlink()
    cmd: tuple[str, ...] = (
        'update-desktop-database',
        *(('-v',) if debug else ()),
        str(Path.home() / '.local/share/applications'),
    )
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)
    cmd = (
        'update-mime-database',
        *(('-v',) if debug else ()),
        str(Path.home() / '.local/share/mime'),
    )
    log.debug('Running: %s', ' '.join(quote(x) for x in cmd))
    sp.run(cmd, check=True)


def secure_move_path(
    client: SSHClient,
    filename: StrPath,
    remote_target: str,
    *,
    dry_run: bool = False,
    preserve_stats: bool = False,
    write_into: bool = False,
) -> None:
    """
    Move a file or directory to a remote host over SSH.

    Like ``scp`` but deletes the local file after a successful copy.

    Parameters
    ----------
    client : paramiko.SSHClient
        Connected SSH client.
    filename : StrPath
        Local file or directory to move.
    remote_target : str
        Remote destination path. ``~`` is expanded to the remote home directory.
    dry_run : bool
        If ``True``, do not perform any file operations.
    preserve_stats : bool
        If ``True``, preserve modification and access times.
    write_into : bool
        If ``True``, write into the target directory rather than creating a new one.
    """
    log.debug('Source: "%s", remote target: "%s"', filename, remote_target)

    def mkdir_ignore_existing(sftp: SFTPClient, td: str, times: tuple[float, float]) -> None:
        if not write_into:
            log.debug('MKDIR "%s"', td)
            if not dry_run:
                sftp.mkdir(td)
                if preserve_stats:
                    sftp.utime(td, times)
            return
        try:
            sftp.stat(td)
        except FileNotFoundError:
            log.debug('MKDIR "%s"', td)
            if not dry_run:
                sftp.mkdir(td)
                if preserve_stats:
                    sftp.utime(td, times)

    path = Path(filename)
    _, stdout, __ = client.exec_command('echo "${HOME}"')
    remote_target = remote_target.replace('~', stdout.read().decode().strip())
    with client.open_sftp() as sftp:
        if path.is_file():
            if not dry_run:
                sftp.put(filename, remote_target)
                if preserve_stats:
                    local_s = Path(filename).stat()
                    sftp.utime(remote_target, (local_s.st_atime, local_s.st_mtime))
            log.debug('Deleting local file "%s".', path)
            if not dry_run:
                path.unlink()
        else:
            pf = Path(filename)
            pf_stat = pf.stat()
            bn_filename = pf.name
            dn_prefix = str(pf).replace(bn_filename, '')
            mkdir_ignore_existing(sftp, remote_target, (pf_stat.st_atime, pf_stat.st_mtime))
            for root, dirs, files in os.walk(filename, followlinks=True):
                p_root = Path(root)
                remote_target_dir = f'{remote_target}/{bn_filename}'
                p_root_stat = p_root.stat()
                mkdir_ignore_existing(sftp, remote_target_dir,
                                      (p_root_stat.st_atime, p_root_stat.st_mtime))
                for name in sorted(dirs):
                    p_root_stat = (p_root / name).stat()
                    dp = str(p_root / name).replace(dn_prefix, '')
                    remote_target_dir = f'{remote_target}/{dp}'
                    mkdir_ignore_existing(sftp, remote_target_dir,
                                          (p_root_stat.st_atime, p_root_stat.st_mtime))
                for name in sorted(files):
                    src = p_root / name
                    dp = str(p_root / name).replace(dn_prefix, '')
                    log.debug('PUT "%s" "%s/%s"', src, remote_target, dp)
                    if not dry_run:
                        sftp.put(src, f'{remote_target}/{dp}')
                        if preserve_stats:
                            local_s = Path(src).stat()
                            sftp.utime(f'{remote_target}/{dp}',
                                       (local_s.st_atime, local_s.st_mtime))
            if not dry_run:
                rmtree(filename, ignore_errors=True)
            else:
                log.debug('Would delete local directory "%s".', filename)


@overload
def kill_processes_by_name(name: str) -> None:  # pragma: no cover
    pass


@overload
def kill_processes_by_name(name: str,
                           wait_timeout: float,
                           signal: int = SIGTERM,
                           *,
                           force: bool = False) -> list[int]:  # pragma: no cover
    pass


def kill_processes_by_name(name: str,
                           wait_timeout: float | None = None,
                           signal: int = SIGTERM,
                           *,
                           force: bool = False) -> list[int] | None:
    """
    Kill processes matching a name.

    On Windows, uses ``taskkill.exe``. On other platforms, uses ``killall``.

    Parameters
    ----------
    name : str
        Process name to kill.
    wait_timeout : float | None
        Seconds to wait after sending the signal. If ``None``, do not wait or collect PIDs.
    signal : int
        Signal number to send.
    force : bool
        If ``True``, forcefully kill remaining processes after the timeout.

    Returns
    -------
    list[int] | None
        List of PIDs that were found after the initial signal, or ``None`` if ``wait_timeout``
        is ``None``.
    """
    name = f'{name}{Path(name).suffix or ".exe"}' if IS_WINDOWS else name
    pids: list[int] = []
    if IS_WINDOWS:
        sp.run(('taskkill.exe', '/im', name), check=False, capture_output=True)
    else:
        sp.run(('killall', f'-{signal}', name), check=False, capture_output=True)
    if wait_timeout:
        lines = sp.run(
            ('tasklist.exe', '/fo', 'csv', '/fi', f'IMAGENAME eq {name}') if IS_WINDOWS else
            ('ps', 'ax'),
            check=True,
            capture_output=True,
            text=True,
        ).stdout.splitlines()
        if (pids := [int(x[1]) for x in list(csv.reader(lines))[1:]] if IS_WINDOWS else
            [int(y[0]) for y in (x.split() for x in lines) if Path(y[0]).name == name]):
            time.sleep(wait_timeout)
            if force:
                sp.run(
                    (
                        'taskkill.exe',
                        *(t for sl in (('/pid', str(pid)) for pid in pids) for t in sl),
                        '/f',
                    ) if IS_WINDOWS else ('kill', '-9', *(str(x) for x in pids)),
                    check=False,
                    capture_output=True,
                )
    return pids if wait_timeout else None


class DataAdapter(BaseAdapter):
    """Requests adapter that returns the URL content (after ``data:``) as the response body."""
    @override
    def send(
        self,
        request: requests.PreparedRequest,
        stream: bool = False,
        timeout: float | tuple[float, float] | tuple[float, None] | None = None,
        verify: bool | str = True,
        cert: bytes | str | tuple[bytes | str, bytes | str] | None = None,
        proxies: Mapping[str, str] | None = None,
    ) -> requests.Response:
        """
        Send a request and return the URL content as the response body.

        Parameters
        ----------
        request : requests.PreparedRequest
            The prepared request.
        stream : bool
            Unused.
        timeout : float | tuple[float, float] | tuple[float, None] | None
            Unused.
        verify : bool | str
            Unused.
        cert : bytes | str | tuple[bytes | str, bytes | str] | None
            Unused.
        proxies : Mapping[str, str] | None
            Unused.

        Returns
        -------
        requests.Response
            A response with the URL content (after ``data:``) as the body.
        """
        r = requests.Response()
        assert request.url is not None
        r._content = request.url[5:].encode()  # noqa: SLF001
        r.status_code = HTTPStatus.OK
        return r

    @override
    def close(self) -> None:
        """Clean up adapter resources."""



from __future__ import annotations

from deltona.naming import Mode, adjust_title


def test_adjust_title_basic_english() -> None:
    assert adjust_title('the quick brown fox') == 'The Quick Brown Fox'
    assert adjust_title('a tale of two cities') == 'A Tale of Two Cities'
    assert adjust_title('in the heat of the night') == 'In the Heat of the Night'


def test_adjust_title_with_stop_words() -> None:
    assert adjust_title('by the river and the sea') == 'By the River and the Sea'
    assert adjust_title('feat john smith') == 'Feat John Smith'
    assert adjust_title('mr smith goes to washington') == 'Mr Smith Goes to Washington'


def test_adjust_title_with_names_to_fix() -> None:
    assert adjust_title("mcdonald's farm") == "McDonald's Farm"
    assert adjust_title('imessage update') == 'iMessage Update'
    assert adjust_title('ios release') == 'iOS Release'
    assert adjust_title('itunes music') == 'iTunes Music'
    assert adjust_title('ios tvos watchos macos imessage') == 'iOS tvOS watchOS macOS iMessage'


def test_adjust_title_with_names_disabled() -> None:
    assert adjust_title("mcdonald's farm", disable_names=True) != "McDonald's Farm"


def test_adjust_title_ampersands() -> None:
    assert adjust_title('rock and roll', ampersands=True) == 'Rock & Roll'
    assert (adjust_title('love and peace and happiness',
                         ampersands=True) == 'Love & Peace & Happiness')


def test_adjust_title_roman_numerals() -> None:
    assert adjust_title('world war ii') == 'World War II'
    assert adjust_title('chapter ix') == 'Chapter IX'
    assert adjust_title('extended mix') == 'Extended Mix'  # 'mix' is ignored for roman numeral


def test_adjust_title_ordinals() -> None:
    assert adjust_title('the 1st time') == 'The 1st Time'
    assert adjust_title('my 2nd home') == 'My 2nd Home'
    assert adjust_title('the 3rd degree') == 'The 3rd Degree'
    assert adjust_title('the 4th wall') == 'The 4th Wall'


def test_adjust_title_with_punctuation() -> None:
    assert adjust_title('hello, world!') == 'Hello, World!'
    assert adjust_title('the end.') == 'The End.'
    assert adjust_title("what's up?") == "What's Up?"


def test_adjust_title_uppercase_preserved() -> None:
    assert adjust_title('NASA launches rocket') == 'NASA Launches Rocket'
    assert adjust_title('WWDC keynote') == 'WWDC Keynote'


def test_adjust_title_multiple_modes() -> None:
    # Should lowercase Japanese particles if Mode.Japanese is used
    result = adjust_title('no wa to', modes=(Mode.Japanese,))
    assert result == 'No wa to'
    # Should lowercase Arabic stops if Mode.Arabic is used
    result = adjust_title('al wa min', modes=(Mode.Arabic,))
    assert result == 'Al wa min'


def test_adjust_title_roman_numeral_at_beginning() -> None:
    assert adjust_title('ii world war I ANOTHER1') == 'II World War I ANOTHER1'
    assert adjust_title('iv       \n    seasons') == 'IV Seasons'


def test_adjust_title_abbreviations() -> None:
    assert (adjust_title('Song Name Of The Year (Feat. Artist Name)') ==
            'Song Name of the Year (feat Artist Name)')



"""Desktop commands."""

from __future__ import annotations

from pathlib import Path
from shlex import quote, split
from shutil import which
from time import sleep
from typing import TYPE_CHECKING, Any, Literal, overload
import contextlib
import errno
import logging
import re
import socket
import subprocess as sp
import webbrowser

from bascom import setup_logging
from deltona.constants import CONTEXT_SETTINGS
from deltona.media import ffprobe
from deltona.string import is_url
from deltona.system import (
    IS_LINUX,
    find_bluetooth_device_info_by_name,
    inhibit_notifications,
    kill_gamescope,
)
from deltona.www import upload_to_imgbb
from platformdirs import user_state_path
from requests import HTTPError
import click
import pyperclip
import requests

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from types import ModuleType

    from deltona.typing import ProbeDict, StreamDispositionDict
    from gi.repository import GLib
    from pydbus.bus import Bus

log = logging.getLogger(__name__)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug logging.')
@click.option(
    '-t',
    '--sleep-time',
    default=0,
    type=int,
    help='Sleep time in seconds to inhibit notifications for.',
)
def inhibit_notifications_main(sleep_time: int = 60, *, debug: bool = False) -> None:
    """
    Disable notifications state for a time.

    On exit, notifications will be enabled. This command does nothing if notifications are already
    disabled.

    This is an alternative to ``kde-inhibit``. Unlike ``kde-inhibit``, this tool may only sleep.
    A sleep time of ``0`` effectively does nothing.
    """
    setup_logging(debug=debug, loggers={'deltona': {}})
    if inhibit_notifications():
        sleep(sleep_time)


@click.command(context_settings={**CONTEXT_SETTINGS, 'auto_envvar_prefix': 'UMPV'})
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('--mpv-command', default='mpv', help='mpv command including arguments.')
@click.argument('files', type=click.Path(path_type=Path), nargs=-1)
def umpv_main(files: Sequence[Path], mpv_command: str = 'mpv', *, debug: bool = False) -> None:
    """Run a single instance of mpv."""  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}})
    fixed_files = ((p if is_url(p) else str(p.resolve(strict=True))) for p in files)
    socket_path = str(user_state_path() / 'umpv-socket')
    sock = None
    socket_connected = False
    try:
        sock = socket.socket(socket.AF_UNIX)
        sock.connect(socket_path)
        socket_connected = True
    except OSError as e:
        match e.errno:
            case errno.ECONNREFUSED:
                log.debug('Socket refused connection')
                sock = None  # abandoned socket
            case errno.ENOENT:
                log.debug('Socket does not exist')
                sock = None  # does not exist
            case _:
                log.exception('Socket errno: %d', e.errno)
                raise click.Abort from e
    if sock and socket_connected:
        # Unhandled race condition: what if mpv is terminating right now?
        for f in fixed_files:
            # escape: \ \n "
            g = str(f).replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
            log.debug('Loading file "%s"', f)
            sock.send(f'raw loadfile "{g}"\n'.encode())
    else:
        log.debug('Starting new mpv instance')
        # Let mpv recreate socket if it does not already exist
        args = (
            *split(mpv_command),
            *(() if debug else ('--no-terminal',)),
            '--force-window',
            f'--input-ipc-server={socket_path}',
            '--',
            *fixed_files,
        )
        log.debug('Command: %s', ' '.join(quote(x) for x in args))
        sp.run(args, check=True)


def _get_pydbus_system_bus_callable() -> Callable[[], Bus]:  # pragma: no cover
    from pydbus import SystemBus  # noqa: PLC0415

    return SystemBus


def _get_gi_repository_glib() -> ModuleType:  # pragma: no cover
    from gi.repository import GLib  # type: ignore[unused-ignore] # noqa: PLC0415

    return GLib


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('--device',
              'device_name',
              default='hci0',
              help='Bluetooth device (defaults to hci0).')
def connect_g603_main(device_name: str = 'hci0', *, debug: bool = False) -> None:
    """
    Connect a G603 Bluetooth mouse, disconnecting/removing first if necessary.

    For Linux only.

    This is useful for connecting the mouse back when it randomly decides not to re-pair, and you
    have no other mouse but you can get to your terminal.
    """  # noqa: DOC501
    if not IS_LINUX:
        click.echo('Only Linux is supported.', err=True)
        raise click.Abort
    from gi.repository import Gio  # noqa: PLC0415

    try:
        g_lib = _get_gi_repository_glib()
        system_bus = _get_pydbus_system_bus_callable()
    except (ImportError, ModuleNotFoundError) as e:
        click.echo('Imports are missing.', err=True)
        raise click.Abort from e
    loop = g_lib.MainLoop()
    setup_logging(debug=debug, loggers={'deltona': {}})
    bus = system_bus()
    adapter = bus.get('org.bluez', f'/org/bluez/{device_name}')

    def on_properties_changed(_: Any, __: Any, object_path: str, ___: Any, ____: Any,
                              props: GLib.Variant) -> None:
        unpacked = props.unpack()
        dev_iface = unpacked[0]
        values = unpacked[1]
        if dev_iface == 'org.bluez.Adapter1' and values.get('Discovering'):
            log.debug('Scan on.')
        elif dev_iface == 'org.bluez.Device1' and (m := re.match(
                rf'/org/bluez/{device_name}/dev_(.*)', object_path)):
            mac = m.group(1).replace('_', ':')
            for k in ('ServicesResolved', 'Connected'):
                if k in values and not values[k]:
                    log.debug('Device %s was disconnected.', mac)
                    return
            try:
                device = bus.get('org.bluez', object_path)
                if device.Name != 'G603':
                    log.debug('Ignoring device %s (MAC: %s).', device.Name, mac)
                    return
            except (KeyError, RuntimeError) as e:
                log.debug('Caught error with device %s: %s', mac, e)
                return
            if values.get('Paired'):
                log.debug('Quitting.')
                loop.quit()
                return
            if 'RSSI' in values:
                click.echo(f'Pairing with {mac}.')
                device['org.bluez.Device1'].Pair()
            else:
                log.debug(
                    'Unhandled property changes: interface=%s, values=%s, mac=%s',
                    dev_iface,
                    values,
                    mac,
                )

    # PropertiesChanged.connect()/.PropertiesChanged = ... will not catch the device node events
    bus.con.signal_subscribe(
        None,
        'org.freedesktop.DBus.Properties',
        'PropertiesChanged',
        None,
        None,
        Gio.DBusSignalFlags.NONE,
        on_properties_changed,
    )
    log.debug('Looking for existing devices.')
    with contextlib.suppress(KeyError):
        while res := find_bluetooth_device_info_by_name('G603'):
            object_path, info = res
            log.debug('Removing device with MAC address %s.',
                      info['Address'])  # type: ignore[typeddict-item]
            adapter.RemoveDevice(object_path)
    click.echo('Put the mouse in pairing mode and be very patient.')
    log.debug('Starting scan.')
    adapter.StartDiscovery()
    try:
        loop.run()
    except KeyboardInterrupt:
        loop.quit()


@click.command(context_settings=CONTEXT_SETTINGS)
def kill_gamescope_main() -> None:
    """Terminate gamescope and gamescopereaper."""
    kill_gamescope()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filenames', type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1)
@click.option('--api-key', help='API key.', metavar='KEY')
@click.option('--keyring-username', help='Keyring username override.', metavar='USERNAME')
@click.option('--no-browser', is_flag=True, help='Do not open browser.')
@click.option('--no-clipboard', is_flag=True, help='Do not copy URL to clipboard.')
@click.option('--no-gui', is_flag=True, help='Disable GUI interactions.')
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
@click.option('-t',
              '--timeout',
              type=float,
              default=5,
              help='Timeout in seconds.',
              metavar='TIMEOUT')
@click.option(
    '--xdg-install',
    default=None,
    metavar='PATH',
    help=('Install .desktop file. Argument is the installation prefix such as /usr. Use '
          '- to install to user XDG directory.'),
)
def upload_to_imgbb_main(
    filenames: Sequence[Path],
    api_key: str | None = None,
    keyring_username: str | None = None,
    timeout: float = 5,
    xdg_install: str | None = None,
    *,
    debug: bool = False,
    no_browser: bool = False,
    no_clipboard: bool = False,
    no_gui: bool = False,
) -> None:
    """
    Upload image to ImgBB.

    Get an API key at https://api.imgbb.com/ and set it with `keyring set imgbb "${USER}"`.
    """  # noqa: DOC501
    setup_logging(debug=debug, loggers={'deltona': {}, 'urllib3': {}})
    r: requests.Response | None = None
    if xdg_install:
        prefix = str(Path('~/.local').expanduser()) if xdg_install == '-' else xdg_install
        apps = Path(f'{prefix}/share/applications')
        apps.mkdir(parents=True, exist_ok=True)
        (apps / 'upload-to-imgbb.desktop').write_text("""[Desktop Entry]
Categories=Graphics;2DGraphics;RasterGraphics;
Exec=upload-to-imgbb %U
Icon=imgbb
Keywords=graphic;design;
MimeType=image/avif;image/gif;image/jpeg;image/png;image/webp
Name=Upload to ImgBB
StartupNotify=false
Terminal=false
TryExec=upload-to-imgbb
Type=Application
Version=1.0
    """)
        r = requests.get('https://simgbb.com/images/favicon.png', timeout=5)
        icons_dir = Path(f'{prefix}/share/icons/hicolor/300x300/apps')
        icons_dir.mkdir(parents=True, exist_ok=True)
        (icons_dir / 'imgbb.png').write_bytes(r.content)
        sp.run(('update-desktop-database', '-v', str(apps)), check=True, capture_output=not debug)
        return
    kdialog = which('kdialog')
    show_gui = not no_gui and len(filenames) == 1 and kdialog
    try:
        for name in filenames:
            r = upload_to_imgbb(name,
                                api_key=api_key,
                                keyring_username=keyring_username,
                                timeout=timeout)
            if not show_gui:
                click.echo(r.json()['data']['url'])
    except HTTPError as e:
        if show_gui:
            assert kdialog is not None
            sp.run((kdialog, '--sorry', 'Failed to upload!'), check=False)
        click.echo('Failed to upload. Check API key!', err=True)
        raise click.Abort from e
    if r:
        url: str = r.json()['data']['url']
        if not no_clipboard:
            pyperclip.copy(url)
        if show_gui:
            click.echo(url)
            assert kdialog is not None
            sp.run((kdialog, '--title', 'Successfully uploaded', '--msgbox', url), check=False)
        elif not no_browser:
            webbrowser.open(url)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('filenames', type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=2)
@click.option('-d', '--debug', is_flag=True, help='Enable debug output.')
def mpv_sbs_main(
    filenames: tuple[Path, Path],
    max_width: int = 3840,
    min_height: int = 31,
    min_width: int = 31,
    *,
    debug: bool = False,
) -> None:
    """Display two videos side by side in mpv."""
    @overload
    def get_prop(prop: Literal['codec_type'],
                 info: ProbeDict) -> Literal['audio', 'video']:  # pragma: no cover
        ...

    @overload
    def get_prop(prop: Literal['disposition'],
                 info: ProbeDict) -> StreamDispositionDict:  # pragma: no cover
        ...

    @overload
    def get_prop(prop: Literal['height', 'width'], info: ProbeDict) -> int:  # pragma: no cover
        ...

    def get_prop(prop: Literal['codec_type', 'disposition', 'height', 'width'],
                 info: ProbeDict) -> Literal['audio', 'video'] | StreamDispositionDict | int:
        return max(
            (x for x in info['streams'] if x['codec_type'] == 'video'),
            key=lambda x: x['disposition'].get('default', 0),
        )[prop]

    def get_default_video_index(info: ProbeDict) -> int:
        return next(
            (i
             for i, x in enumerate(info['streams']) if x.get('disposition', {}).get('default', 0)),
            next((i for i, x in enumerate(info['streams']) if x['codec_type'] == 'video'), 0),
        )

    setup_logging(debug=debug, loggers={'deltona': {}})
    filename1, filename2 = filenames
    info1, info2 = ffprobe(filename1), ffprobe(filename2)
    width1, width2 = (int(get_prop('width', info1)), int(get_prop('width', info2)))
    height1, height2 = (int(get_prop('height', info1)), int(get_prop('height', info2)))
    assert height1 > min_height, 'Invalid height in video 1'
    assert height2 > min_height, 'Invalid height in video 2'
    assert width1 <= max_width, 'Video 1 is too wide'
    assert width1 > min_width, 'Invalid width in video 1'
    assert width2 <= max_width, 'Video 2 is too wide'
    assert width2 > min_width, 'Invalid width in video 2'
    scale_w = max(width1, width2)
    scale_h = (int(get_prop('height', info1)) if scale_w == width1 else int(
        get_prop('height', info2)))
    scale = '' if width1 == width2 and height1 == height2 else f'scale={scale_w}x{scale_h}'
    scale1, scale2 = scale if scale_h != height1 else '', scale if scale_h == height1 else ''
    second_stream_index = (len([x for x in info1['streams'] if x['codec_type'] == 'video']) +
                           get_default_video_index(info2)) + 1
    if not scale1 and not scale2:
        filter_chain = '[vid1][vid2] hstack [vo]'
    else:
        filter_chain = ';'.join((
            f'[vid1] {scale} [vid1_scale]',
            f'[vid{second_stream_index}] {scale} [vid{second_stream_index}_crop]',
            f'[vid1_scale][vid{second_stream_index}_crop] hstack [vo]',
        ))
    cmd = (
        'mpv',
        '--hwdec=no',
        '--config=no',
        str(filename1),
        f'--external-file={filename2}',
        f'--lavfi-complex={filter_chain}',
    )
    log.debug('Running: %s', ' '.join(quote(str(x)) for x in cmd))
    sp.run(cmd, check=True)
